#     Copyright (C) 2006
#
#     This code is distributed under the terms and conditions of the
#     CCP4 Program Suite Licence Agreement as a CCP4 Library.
#     A copy of the CCP4 licence can be obtained by writing to the
#     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
#
#====================================================================
# CCP4 Interface - dbClientAPI.py
#
# Classes and methods for interacting with the database handler
# 
# Wanjuan Yang and Peter Briggs
#
#====================================================================

#CCP4i_cvs_Id $Id: dbClientAPI.py,v 1.25 2007/02/14 16:28:39 wy45 Exp $

##################################################################

#
#d_index_title Database Handler Client API (ClientAPI/dbClientAPI.py)
#d_intro_title Functions for applications interacting with the DB server
#
#d_intro These are the commands that can be invoked by a client of \
#the database server in order to communicate with it. The commands map onto the commands provided in the database server.

import socket
import exceptions
import sys
import os
import re
import time
import ccp4i
import dbsocket

def DbStartHandler():
    """Start a dbccp4i handler process in background."""
    
    handler = os.path.join(ccp4i.GetEnvVar("BIOXHIT_TOP"), \
                           "dbccp4i","dbccp4i","dbccp4i.py")
    os.system("python "+str(handler)+" &")

def GetHandlerPort(nowait=False):
    """Acquire the port number for a handler process.

    Attempts to read the port number from the handler lock file in the
    user's .CCP4 area, and returns the port number if successful (otherwise
    returns -1). Set nowait=True to prevent the function from making
    multiple attempts to get the port number before giving up."""

    # The lockfile is called dbccp4i.LOCK and has the following content:
    # port number is:4090
    if not nowait:
        ntries = 10
        wait_time = 1
    else:
        ntries = 1
        wait_time = 0
    thistry = 1
    port = -1
    rlock = re.compile(r"^port number is:([0-9]+)$")
    lockfile = os.path.join(ccp4i.GetDotCCP4(),"dbccp4i.LOCK")
    while thistry < ntries:
        print "GetHandlerPort: looking for lockfile"
        if ccp4i.FileExists(lockfile):
            # Open and read the contents
            print "GetHandlerPort: opening lockfile"
            fp = open(lockfile,"r")
            line = fp.readline().rstrip()
            print "GetHandlerPort: line from lockfile is "+str(line)
            if rlock.match(line):
                port = int(rlock.match(line).group(1))
            fp.close()
            print "GetHandlerPort: port "+str(port)
            # Try to connect to the port and request the
            # server status
            try:
                addr = ('localhost',port)
                sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
                sock.connect(addr)
                request = dbsocket.dbRequest('DbRequestStatus',[])
                sock.send(request.xml()+"\n")
                xml_response = sock.recv(4096)
                response = dbsocket.dbResponse(xml_response)
                print "GetHandlerPort: response from handler: " \
                      +str(response.result())
                request = dbsocket.dbRequest('DbDisconnect',[])
                sock.send(request.xml()+"\n")
                sock.close()
                # Check the response
                if str(response.result()) == 'Active':
                    # Handler is alive - return the port number
                    return port
            except exceptions.Exception,e:
                print "GetHandlerPort: connection failed with exception:"
                print str(e)
                return -1
        # Wait for a little while then retry
        time.sleep(wait_time)
        thistry = thistry + 1
    # Timeout
    print "GetHandlerPort: timed out waiting for lockfile"
    return -1

def ConvertTime(secs):
    """ Given a time expressed in seconds, convert into formated time """
    return time.strftime("%d %b %Y  %H:%M:%S",time.localtime(secs))

class handlerconnection:

    """ create a connection to the handler, provide commands to talk to handler, access database.def file & directories.def. """

    def __init__(self):
        self.host = 'localhost'
        self.port = GetHandlerPort()
        self.sock = dbsocket.dbsocket(self.host,self.port)
    
    def DbRegister(self,user,pid,broadcast=True):
        """ Register with handler """
        argu_list = [user,pid,broadcast]
        return self.sock.request('DbRegister',argu_list)
        
    def CheckServerStatus(self):
        """ Return handler status """
        argu_list = []
        return self.sock.request('DbRequestStatus',argu_list)
        
    def OpenDatabase(self,project,grablock=False):
        """ Open an existing project """
        argu_list = [project,grablock]
        return self.sock.request('DbOpen',argu_list)
    

    def CreateDatabase(self,project,dir):
        """ Create a new project """
        argu_list = [project,dir]
        return self.sock.request('NewDb',argu_list)

    def Updatetime(self,project,jobid):
        """ Update the time of a job """
        argu_list = [project,jobid]
        return self.sock.request('Updatetime',argu_list)

    def ListJobs(self,project):
        """ List all of the jobs in a project """
        argu_list = [project]
        return self.sock.request('ListJobs',argu_list)

    def ImportProject(self,alias,path):
        """ Import a project """
        argu_list = [alias,path]
        return self.sock.request('ImportProject',argu_list)

    def DeleteProject(self,project):
        """ Delete a project """
        argu_list = [project]
        return self.sock.request('DeleteProject',argu_list)

    def GetNextJobList(self,project,jobid):
        """ Return all the jobs that link to the job """
        argu_list = [project,jobid]
        return self.sock.request('GetNextJobList',argu_list)

    def GetAllLinks(self,project):
        """ Return all the links between jobs in a project """
        argu_list = [project]
        return self.sock.request('GetAllLinks',argu_list)

    def GetLinks(self,project,joblist):
        """ Return the links for specified jobs """
        argu_list = [project,joblist]
        return self.sock.request('GetLinks',argu_list)

    def GetAllChildren(self,project,jobid):
        """ Return all the descendents of a job  """
        argu_list = [project,jobid]
        return self.sock.request('GetAllChildren',argu_list)

    def GetAllParents(self,project,jobid):
        """ Return all the antecedents of a job """
        argu_list = [project,jobid]
        return self.sock.request('GetAllParents',argu_list)

    def GetParents(self,project,jobid):
        """ Return the immediate parents of the job """
        argu_list = [project,jobid]
        return self.sock.request('GetParents',argu_list)
    
    def GetChildren(self,project,jobid):
        """ Return the immediate children of the job """
        argu_list = [project,jobid]
        return self.sock.request('GetChildren',argu_list)
    
    def GetListofRecords(self,project,joblist,itemlist):
        """ Return a list of records for a project """
        argu_list = [project,joblist,itemlist]
        return self.sock.request('DbGetListofRecords',argu_list)
    
    def GetFileList(self,project,jobid,type):
        """ Return a list of files associated with a job. type can be 'INPUT','OUTPUT' """
        # Get lists of project and defdirs
        projectlist = self.ListProjects()[1]
        defdirlist = self.ListDefDirs()[1]
       
        # Get job file name & file dir
        itemlist = [str(type)+"_FILES",str(type)+"_FILES_DIR"]
        records = self.GetListofRecords(project,[jobid],itemlist)

        # get the files & files dir
        files = records[1][0][0]
        files_list = files.split(' ')
        file_dir = records[1][0][1]
        file_dir_list = file_dir.split(' ')

        filelist = []
       
        for i in range(len(files_list)):
            filen = files_list[i]
     
            if filen == "{}":
                filen = ""
            alias = file_dir_list[i]
          
            if alias == "FULL_PATH" or alias == "{}":
                dir = ""
            else:
                # from alias get the dir
                if projectlist.count(alias) == 1:
                    dir = self.GetProjectDir(alias)[1]
                elif defdirlist.count(alias) == 1:
                    dir = self.GetDefDir(alias)[1]
                else:
                    dir = ""
            if dir == "":
                full_filen = filen
            else:
                # assemble the full filename
                full_filen = os.path.join(dir,filen)
           
            filelist.append(full_filen)

        return filelist        
        
    def HandlerDisconnect(self):
        """ Disconnect from handler """
        argu_list = []
        self.sock.request('DbDisconnect',argu_list)
        try:
            self.sock.close()
        except:
            print "Exception in HandlerDisconnect"
        
    def ShutDown(self):
        """ Shut down the handler """
        argu_list = []
        self.sock.request('Shutdown',argu_list)
        self.sock.close()

    def CloseDatabase(self,project):
        """ Close a project """
        argu_list = [project]
        return self.sock.request('DbClose',argu_list)
    
    def NewRecord(self,project):
        """ Create a job in a project """
        argu_list = [project]
        return self.sock.request('DbNewRecord',argu_list)

    def DeleteRecord(self,project,job_id):
        """ Delete a job in a project """
        argu_list = [project,job_id]
        return self.sock.request('DbDeleteRecord',argu_list)

    def ItemExists(self,project,job_id,item):
        """ Check if an item exist for a project """
        argu_list = [project,job_id,item]
        return self.sock.request('DbItemExists',argu_list)

    def SetData(self,project,job_id,item,value):
        """ Set item value for a job in a project """
        argu_list = [project,job_id,item,value]
        return self.sock.request('DbSetData',argu_list)

    def GetData(self,project,job_id,item):
        """ Return item value for a job in a project """
        argu_list = [project,job_id,item]
        return self.sock.request('DbGetData',argu_list)
 
    def GetDescription(self,project,job_list,db_display,db_display_format):
        """ Return the description of a job in a project """
        argu_list = [project,job_list,db_display,db_display_format]
        return self.sock.request('DbReturnRecord',argu_list)
    
    def SelectJobs(self,project,item,pattern):
        """ Retrieve a list of jobs based on some selection criterion """
        argu_list = [project,item,pattern]
        return self.sock.request('DbSelectRecords',argu_list)

    def ListProjects(self):
        """ List all the user's projects """
        argu_list = []
        return self.sock.request('ListProjects',argu_list)

    def DeleteProject(self,project):
        """ Delete a project reference """
        argu_list = [project]
        return self.sock.request('DeleteProject',argu_list)

    def GetProjectDir(self,project):
        """ Return the directory corresponding to a project name """
        argu_list = [project]
        return self.sock.request('GetProjectDir',argu_list)

    def ListDefDirs(self):
        """ Return a list of the user's default directories """
        argu_list = []
        return self.sock.request('ListDefDirs',argu_list)

    def GetDefDir(self,def_dir):
        """  Return the directory corresponding to a def dir name """
        argu_list = [def_dir]
        return self.sock.request('GetDefDir',argu_list)

    def GetNJOB(self,project):
        """ Return the highest job number in the project """
        argu_list = [project]
        return self.sock.request('GetNJOB',argu_list)

    def GetNProjects(self):
        """ Return the number of projects """
        argu_list = []
        return self.sock.request('GetNProjects',argu_list)

    def OpenDBsql(self,dbname,dir):
        """ Open a SQLite database """
        argu_list = [dbname,dir]
        return self.sock.request('OpenDB_sql',argu_list)

    def NewDBsql(self,dbname,dir):
        """ Create a SQLite database """
        argu_list = [dbname,dir]
        return self.sock.request('NewDB_sql',argu_list)
    
    def NewProjectsql(self,projectname):
        """ Create a project in SQLite database """
        argu_list = [projectname]
        return self.sock.request('NewProject_sql',argu_list)

    def NewJobsql(self,projectname):
        """ Create a job in SQLite database """
        argu_list = [projectname]
        return self.sock.request('NewJob_sql',argu_list)

    def SetJobDatasql(self,job_id,item,value):
        """ Set job data in SQLite database """
        argu_list = [job_id,item,value]
        return self.sock.request('SetJobData_sql',argu_list)
    
    def GetJobDatasql(self,job_id,item):
        """ Return job data in SQLite database """
        argu_list = [job_id,item]
        return self.sock.request('GetJobData_sql',argu_list)

    def CloseDBsql(self):
        """ Close a SQLite database """
        argu_list = []
        return self.sock.request('CloseDB_sql',argu_list)

    def BroadcastReporter(self):
        """ Checks for broadcast messages and print them to stdout """
        dbsocket.report_broadcast(self.sock)


class db:
    """  create a SQLite db object """
    def __init__(self,con):
        self.conn = con

    def sendrequest(self,command,argu_list):
        """ send request to handler and return response """
        response = self.conn.sock.request('',argu_list)
        return parse_response(response)

    def OpenDB(self,dbname,dir):
        """ Open a database """
        argu_list = [dbname,dir]
        return self.sendrequest('OpenDB_sql',argu_list)

    def NewDB(self,dbname,dir):
        """ Create a database """
        argu_list = [dbname,dir]
        return self.sendrequest('NewDB_sql',argu_list)
    
    def CloseDB(self):
        """ Close a database  """
        argu_list = []
        return self.sendrequest('CloseDB_sql',argu_list)

    def NewRecord(self,table):
        """ New a record in a given table """
        argu_list = [table]
        return self.sendrequest('NewRecord',argu_list)
    
    def SetData(self,table,id,attribute,value):
        """ Set attribute value in a table's record """
        argu_list = [table,id,attribute,value]
        return self.sendrequest('SetData',argu_list)

    def GetData(self,table,id,attribute):
        """ Return the attribute value in a table """
        argu_list = [table,id,attribute]
        return self.sendrequest('GetData',argu_list)

    def GetProjectList(self):
        """ Return a project list """
        argu_list = []
        return self.sendrequest('GetProjectList',argu_list)

    def GetJobList(self,projectname):
        """ Return joblist in a project """
        argu_list = [projectname]
        return self.sendrequest('GetJobList',argu_list)
    
    def GetRowofTable(self,table,rid):
        """ Return a record in a table """
        argu_list = [table,rid]
        return self.sendrequest('GetRowofTable',argu_list)

    
class Project:
    """ create a project object, have methods to set and get project attributes """

    def __init__(self,db):
        self.db = db

    def newProject(self,projectname):
        argu_list = [projectname]
        return self.db.sendrequest('NewProject_sql',argu_list)
    
    def setOwner(self,pid,owner):
        return self.db.SetData('Project',pid,'Owner',owner)
      
    def setPath(self,pid,path):
        return self.db.SetData('Project',pid,'Path',path)
      
    def getProjectId(self,projectname):
        argu_list = [projectname]
        return self.db.sendrequest('GetProjectId',argu_list)

    def getOwner(self,pid):
        return self.db.GetData('Project',pid,'Owner')

    def getPath(self,pid):
        return self.db.GetData('Project',pid,'Path')

class Job:

    """ create a job object. have methods of set & get job attributes """
    
    def __init__(self,db):
        self.db = db

    def newJob(self,projectname):
        argu_list = [projectname]
        return self.db.sendrequest('NewJob_sql',argu_list)
    
    def setData(self,jobid,item,value):
        argu_list = [jobid,item,value]
        return self.db.sendrequest('SetData_sql',argu_list)
    
    def getData(self,jobid,item):
        argu_list = [jobid,item]
        return self.db.sendrequest('GetData_sql',argu_list)

    def setStatus(self,jobid,status):
        argu_list = ['Job',jobid,'Status',status]
        return self.db.sendrequest('SetData',argu_list)
        
    def getStatus(self,jobid):
        argu_list=['Job',jobid,'Status']
        return self.db.sendrequest('GetData',argu_list)

    def setApplication(self,JobId,Application):
        argu_list = ['Job',JobId,'Application',Application]
        return self.db.sendrequest('SetData',argu_list)

    def getApplication(self,JobId):
        argu_list = ['Job',JobId,'Application']
        return self.db.sendrequest('GetData',argu_list)

    def setTaskname(self,jobid,taskname):
        argu_list = ['Job',jobid,'Taskname',taskname]
        return self.db.sendrequest('SetData',argu_list)

    def getTaskname(self,jobid):
        argu_list = ['Job',jobid,'Taskname']
        return self.db.sendrequest('GetData',argu_list)

    def setTitle(self,jobid,title):
        argu_list = ['Job',jobid,'Title',title]
        return self.db.sendrequest('SetData',argu_list)

    def getTitle(self,jobid):
        argu_list = ['Job',jobid,'Title']
        return self.db.sendrequest('GetData',argu_list)
    
    def setLastModified(self,jobid,time):
        argu_list = ['Job',jobid,'LastModified',time]
        return self.db.sendrequest('SetData',argu_list)

    def getLastModified(self,jobid):
        argu_list = ['Job',jobid,'LastModified']
        return self.db.sendrequest('GetData',argu_list)

    def setLogFile(self,jobid,logfile):
        argu_list = ['Job',jobid,'LogFile',logfile]
        return self.db.sendrequest('SetData',argu_list)

    def getLogFile(self,jobid):
        argu_list = ['Job',jobid,'LogFile']
        return self.db.sendrequest('GetData',argu_list)

    def setControlFile(self,JobId,ControlFile):
        argu_list = ['Job',JobId,'ControlFile',ControlFile]
        return self.db.sendrequest('SetData',argu_list)

    def getControlFile(self,JobId):
        argu_list = ['Job',JobId,'ControlFile']
        return self.db.sendrequest('GetData',argu_list)

    def setNotebookFile(self,JobId,NotebookFile):
        argu_list = ['Job',JobId,'NotebookFile',NotebookFile]
        return self.db.sendrequest('SetData',argu_list)

    def getNotebookFile(self,JobId):
        argu_list = ['Job',JobId,'NotebookFile']
        return self.db.sendrequest('GetData',argu_list)


class File:

    """ create a file object. have methods set & get file attributes """

    def __init__(self,db):
        self.db = db
 
    def newFile(self):
        argu_list = ['File']
        return self.db.sendrequest('NewRecord',argu_list)

    def setName(self,FileId,Name):
        argu_list = ['File',FileId,'Name',Name]
        return self.db.sendrequest('SetData',argu_list)

    def getName(self,FileId):
        argu_list = ['File',FileId,'Name']
        return self.db.sendrequest('GetData',argu_list)
    
    def setPath(self,FileId,Path):
        argu_list = ['File',FileId,'Path',Path]
        return self.db.sendrequest('SetData',argu_list)

    def getPath(self,FileId):
        argu_list = ['File',FileId,'Path']
        return self.db.sendrequest('GetData',argu_list)

    def setFormat(self,FileId,Format):
        argu_list = ['File',FileId,'Format',Format]
        return self.db.sendrequest('SetData',argu_list)

    def getFormat(self,FileId):
        argu_list = ['File',FileId,'Format']
        return self.db.sendrequest('GetData',argu_list)
   
    def setLastModified(self,FileId,LastModified):
        argu_list = ['File',FileId,'LastModified',LastModified]
        return self.db.sendrequest('SetData',argu_list)

    def getLastModified(self,FileId):
        argu_list = ['File',FileId,'LastModified']
        return self.db.sendrequest('GetData',argu_list)

    def setNote(self,FileId,Note):
        argu_list = ['File',FileId,'Note',Note]
        return self.db.sendrequest('SetData',argu_list)

    def getNote(self,FileId):
        argu_list = ['File',FileId,'Note']
        return self.db.sendrequest('GetData',argu_list)


class JobFile:
    """ create a JobFile object, have methods set & get JobFile attributes """
 
    def __init__(self,db):
        self.db = db
 
    def newJobFile(self):
        argu_list = ['JobFile']
        return self.db.sendrequest('NewRecord',argu_list)

    def setFileId(self,JobFileId,FileId):
        argu_list = ['JobFile',JobFileId,'FileId',FileId]
        return self.db.sendrequest('SetData',argu_list)

    def getFileId(self,JobFileId):
        argu_list = ['JobFile',JobFileId,'FileId']
        return self.db.sendrequest('GetData',argu_list)

    def setType(self,JobFileId,Type):
        argu_list = ['JobFile',JobFileId,'Type',Type]
        return self.db.sendrequest('SetData',argu_list)

    def getType(self,JobFileId):
        argu_list = ['JobFile',JobFileId,'Type']
        return self.db.sendrequest('GetData',argu_list)

class JobLink:
    """ create a JobLink object, have methods set & get JobLink attributes """
    
    def __init__(self,db):
        self.db = db
 
    def newJobLink(self):
        argu_list = ['JobLink']
        return self.db.sendrequest('NewRecord',argu_list)

    def setNextJobId(self,JobLinkId,NextJobId):
        argu_list = ['JobLink',JobLinkId,'NextJobId',NextJobId]
        return self.db.sendrequest('SetData',argu_list)
    
    def getNextJobId(self,JobLinkId):
        argu_list = ['JobLink',JobLinkId,'NextJobId']
        return self.db.sendrequest('GetData',argu_list)

    def setNote(self,JobLinkId,Note):
        argu_list = ['JobLink',JobLinkId,'Note',Note]
        return self.db.sendrequest('SetData',argu_list)

    def getNote(self,JobLinkId):
        argu_list = ['JobLink',JobLinkId,'Note']
        return self.db.sendrequest('GetData',argu_list)

    def setType(self,JobLinkId,Type):
        argu_list = ['JobLink',JobLinkId,'Type',Type]
        return self.db.sendrequest('SetData',argu_list)
    
    def getType(self,JobLinkId):
        argu_list = ['JobLink',JobLinkId,'Type']
        return self.db.sendrequest('GetData',argu_list)
    
