#
#
#     Copyright (C) 2005-2006
#
#     This code is distributed under the terms and conditions of the
#     CCP4 Program Suite Licence Agreement as a CCP4 Library.
#     A copy of the CCP4 licence can be obtained by writing to the
#     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
#
#====================================================================
# CCP4 Interface - dbClientAPI.tcl
#
# API functions to allow a client application to communicate with
# the database handler
# 
# Peter Briggs and Wanjuan Yang
# Based on code originally written by Liz Potterton
#====================================================================

#CCP4i_cvs_Id $Id: dbClientAPI.tcl,v 1.19 2007/02/06 16:17:05 pjx Exp $

##################################################################

#
#d_index_title Database Handler Client API (src/dbClientAPI.tcl)
#d_intro_title Functions for applications interacting with the DB server
#
#d_intro These are the commands that can be invoked by a client of \
the database server in order to communicate with it. The commands map \
onto the commands provided in the database server.

# Global parameters
# PJX these must be added to the appropriate places eventually

set handler(RUN_DBSERVER) "python [file join $env(BIOXHIT_TOP) dbccp4i dbccp4i dbccp4i.py]"

set handler(DB_PORT) ""
set handler(DB_SOCKET) ""
set handler(DB_BROADCAST_CALLBACK) {}
set handler(DB_HANDLER_STATUS) 0
set handler(DB_CONNECTION_STATUS) 0
set handler(DB_TIMEOUT) 5000

# Load the pseudo-XML handling functions
#source [SearchPath TOP utils dbxml.tcl]
source [file join $env(BIOXHIT_TOP) dbccp4i ClientAPI dbxml.tcl]
# This file contains the database primitives
# The primitives hide the implementation details i.e. nothing about
# the handler process should be visible from outside this file.

#-----------------------------------------------------------------------
proc ListProjects { } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to list all projects

  if { [regexp ^OK [DbHandlerRequest message "ListProjects" ]] } {
      return $message
  } else {
      Report 1 "ERROR list projects: $message"
      return 0
  }
}

#-----------------------------------------------------------------------
proc GetNProjects { } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to list all projects

  if { [regexp ^OK [DbHandlerRequest message "GetNProjects" ]] } {
      return $message
  } else {
      Report 1 "ERROR list projects: $message"
      return 0
  }
}


#-----------------------------------------------------------------------
proc ListDefDirs { } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to list user's default directories

  if { [regexp ^OK [DbHandlerRequest message "ListDefDirs" ]] } {
      return $message
  } else {
      Report 1 "ERROR list def dirs: $message"
      return 0
  }
}

#-----------------------------------------------------------------------
proc DeleteProject { project } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to delete a project reference
#d_arg project Alias for the prpject being deleted

  if { [regexp ^OK [DbHandlerRequest message "DeleteProject" $project ]] } {
      return 1
  } else {
      Report 1 "ERROR delete project: $message"
      return 0
  }
}


#-----------------------------------------------------------------------
proc ImportProject { alias path } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to add a project reference
#d_arg project Alias for the prpject being added

  if { [regexp ^OK [DbHandlerRequest message "ImportProject" $alias $path ]] } {
      return 1
  } else {
      Report 1 "ERROR import project: $message"
      return 0
  }
}

#-----------------------------------------------------------------------
proc GetProjectDir { project } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to get a project's directory name
#d_arg project Alias for the project which the directory contains


  if { [regexp ^OK [DbHandlerRequest message "GetProjectDir" $project ]] } {
      return $message
  } else {
      Report 1 "ERROR Get project dir: $message"
      return 0
  }
}


#-----------------------------------------------------------------------
proc GetDefDir { def_dir } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to get the directory corresponding to a default directory name
#d_arg def_dir the default directory name

  if { [regexp ^OK [DbHandlerRequest message "GetDefDir" $def_dir ]] } {
      return $message
  } else {
      Report 1 "ERROR Get def dir: $message"
      return 0
  }
}




#-----------------------------------------------------------------------
proc GetNJOB { project } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to get the NJOB of the database

  if { [regexp ^OK [DbHandlerRequest NJOB "GetNJOB" $project]] } {
      return $NJOB
  } else {
      Report 1 "ERROR GetNJOB: $NJOB"
      return 0
  }
}


#-----------------------------------------------------------------------
proc OpenDatabase { project args } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to open a project database
#d_desc If the first argument doesn't start with a hyphen then it's \
assumed to be the name of a variable in the calling procedure in which \
to report any message returned from the handler. All other options \
must come after this and must start with a hyphen.
#d_desc Return 1 on success and 0 on failure.
#d_arg project Alias for the project being opened
#d_opt messageVar (Optional) Name of a variable to return any messages in
#d_opt -grablock Forces the handler to open a locked database
#d_opt -force An alias for -grablock

  # Process the command arguments to see if a variable name
  # was supplied as the first argument
  set i 0
  set messageVar ""
  if { [llength $args] > 0 } {
      set opt [lindex $args $i]
      if { ![regexp -- "^-(.*)\$" $opt] } {
	  set messageVar $opt
	  incr i
      }
  }
  if { $messageVar != "" } {
      upvar $messageVar message
  }

  # Pass the remaining arguments to the handler request
  # mechanism
  set new_args [lrange $args $i end]

  # Send the request and convert the return value
  set status [eval DbHandlerRequest message "DbOpen" $project $new_args]
  return [regexp -- "^OK\$" $status]
}

#-----------------------------------------------------------------------
proc CreateDatabase { project db messageVar args } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to create a project database
#d_arg project Alias for the project being created
#d_arg db The full directory path of the project database
#d_arg messageVar Name of a variable to return any messages in
    upvar $messageVar message
    return [eval DbHandlerRequest message "NewDb" $project $db $args]
}


#-----------------------------------------------------------------------
proc CloseDatabase { project } {
#-----------------------------------------------------------------------
#d_sum Close a currently open project database
#d_desc The client must previously have opened the database using \
the OpenDatabase command, otherwise an error will be returned.
#d_arg project Alias for the project being closed
  if { [regexp ^OK [DbHandlerRequest message "DbClose" $project]] } {
      return 1
  } else {
      Report 1 "ERROR closing database: $message"
      puts "ERROR closing database: $message"
      return 0
  }
}


#-----------------------------------------------------------------------
proc ListJobs { project } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to list jobs of a given project

  if { [regexp ^OK [DbHandlerRequest message "ListJobs" $project ]] } {
      return $message
  } else {
      Report 1 "ERROR list jobs: $message"
      return 0
  }
}


#-----------------------------------------------------------------------
proc GetAllLinks { project } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to get a project's history link between jobs

  if { [regexp ^OK [DbHandlerRequest message "GetAllLinks" $project ]] } {
      return $message
  } else {
      Report 1 "ERROR: $message"
      return 0
  }
}


#-----------------------------------------------------------------------
proc GetLinks { project jobs } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to get a project's history link between jobs

  if { [regexp ^OK [DbHandlerRequest message "GetLinks" $project -list $jobs]] } {
      return $message
  } else {
      Report 1 "ERROR: $message"
      return 0
  }
}

#-----------------------------------------------------------------------
proc GetNextJobList { project jobid } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to a job's next job list of a given project

  if { [regexp ^OK [DbHandlerRequest message "GetNextJobList" $project $jobid]] } {
      return $message
  } else {
      Report 1 "ERROR list jobs: $message"
      return 0
  }
}


#-----------------------------------------------------------------------
proc GetAllChildren { project jobid } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to a job's next job list of a given project

  if { [regexp ^OK [DbHandlerRequest message "GetAllChildren" $project $jobid]] } {
      return $message
  } else {
      Report 1 "ERROR list jobs: $message"
      return 0
  }
}

#-----------------------------------------------------------------------
proc GetAllParents { project jobid } {
#-----------------------------------------------------------------------
#d_sum Ask the database handler to a job's next job list of a given project

  if { [regexp ^OK [DbHandlerRequest message "GetAllParents" $project $jobid]] } {
      return $message
  } else {
      Report 1 "ERROR list jobs: $message"
      return 0
  }
}




#-----------------------------------------------------------------------
proc NewRecord { project } {
#-----------------------------------------------------------------------
#d_sum Create a new record in a CCP4i project database
#d_desc Creates a new job in the specified job database and returns \
the job id.
#d_arg project Alias for the project in which the job is being created
  if { [regexp ^OK$ [DbHandlerRequest njobs "DbNewRecord" $project ]] } {
    return $njobs
  } else {
    return 0
  }
}

#-----------------------------------------------------------------------
proc DeleteRecord { project job_id } {
#-----------------------------------------------------------------------
#d_sum Delete a record (job) from a CCP4i project database
#d_desc Removes the specified job id and associated data from the \
specified project database.
#d_arg project Alias for the project from which the job is being removed
#d_arg job_id  Number of the job being removed
  if { [regexp ^OK$ \
	  [DbHandlerRequest message "DbDeleteRecord" $project $job_id] ] } {
      return 1
  }
  Report 1 "ERROR deleting record: $message"
  return 0
}

#-----------------------------------------------------------------------
proc ItemExists { project job_id item } {
#-----------------------------------------------------------------------
#d_sum Test for the existence of a data item
#d_desc Returns 1 if the specified data item exists for the named job \
and project database, and 0 if not.
#d_arg project Alias of the project
#d_arg job_id  Number of the job being interrogated
#d_arg item    Name of the data item being queried
  set result [DbHandlerRequest status "DbItemExists" $project $job_id $item]
  Report 1 "ItemExists: result = $result"
  if { ![regexp "^OK\$" $result] } {
      Report 1 "ItemExists: error: $result $status"
      set status 0
  } else {
      # DbItemExists returns True or False
      if { $status == "True" } {
	  set status 1
      } else {
	  set status 0
      }
  }
  return $status
}

#-----------------------------------------------------------------------
proc SetData { project job_id args } {
#-----------------------------------------------------------------------
#d_sum Set the values of one or more data items
#d_desc This sets the values of multiple data items for a particular job \
in the specified project database. The argument list should consist of \
key-value pairs.
#d_desc Setting multiple data items in a single SetData command is \
recommended as this minimises the number of socket transactions required \
to send the data to the server.
#d_arg project Alias of the project database
#d_arg job_id  Number of the job for which the data is being set
#d_arg item1   Item name to set the value of
#d_arg value1  Corresponding value to set "item1" to be
  # Use eval to invoke the command, to stops the arglist concatenated
  # into a single argument
  if { ![regexp ^OK$ \
	  [eval DbHandlerRequest message "DbSetData" \
	       $project $job_id $args]] } {
      Report 1 "ERROR Updating database: $message"
      return 0
  }
  return 1
}

#-----------------------------------------------------------------------
proc GetData { project job_id args } {
#-----------------------------------------------------------------------
#d_sum Return the values of one or more data items
#d_desc This returns the values of multiple data items for a particular \
job in the specified project database. The argument list should consist \
of keys. A list of the corresponding values will be returned.
#d_desc Retrieving multiple data items in a single GetData command is \
recommended as this minimises the number of socket transactions required \
to fetch the data from the server.
#d_arg project Alias of the project database
#d_arg job_id  Number of the job for which the data is being fetched
#d_arg item1   Item name to get the value of
  # Use eval to invoke the command, to stops the arglist concatenated
  # into a single argument
  if { [regexp ^OK$ \
	  [eval DbHandlerRequest value "DbGetData" \
	       $project $job_id $args]] } {
      return $value
  }
  Report 1 "ERROR Extracting data: $value"
  return ""
}



#-----------------------------------------------------------------------
proc GetListofRecords { project joblist itemlist } {
#-----------------------------------------------------------------------
#d_sum Return a list of records
#d_desc This returns the values of multiple data items for a list of given jobs in the specified project database. The argument list should consist of keys. A list of the corresponding values will be returned.
#d_desc Retrieving multiple data items of multiple jobs in a single GetData command is recommended as this minimises the number of socket transactions required to fetch the data from the server.
#d_arg project Alias of the project database
#d_arg job_id  Number of the job for which the data is being fetched
#d_arg item1   Item name to get the value of
  # Use eval to invoke the command, to stops the arglist concatenated
  # into a single argument
  if { [regexp ^OK$ \
	  [DbHandlerRequest value "DbGetListofRecords" \
	       $project -list $joblist -list $itemlist]] } {
      return $value
  }
  Report 1 "ERROR Extracting data: $value"
  return ""
}




#-----------------------------------------------------------------------
proc SelectJobs { project item pattern } {
#-----------------------------------------------------------------------
#d_sum Return a list of jobs matching a regular expression
#d_desc This command returns a list of job numbers where the values of \
the specified item matches the supplied regular expression pattern.
#d_arg project Alias of the project database
#d_arg item    Item name being searched on
#d_arg pattern Regular expression pattern to search on
  if { [regexp ^OK$ \
	  [DbHandlerRequest job_list "DbSelectRecords" \
	       $project $item $pattern]] } {
      return $job_list
  }
  Report 1 "ERROR Extracting data: $job_list"
  return ""
}

#-----------------------------------------------------------------------
proc GetDescription { project job_list db_display db_display_format } {
#-----------------------------------------------------------------------
#d_sum Return a formatted list of jobs in report format
#d_desc This command takes a list of job numbers, a list of item names \
and a format description, and returns a report as a Tcl list with one \
job description (generated according to the lists of items and formats) \
per item.
#d_desc This can be used to append all database information into one \
string for display.
#d_arg project The alias for the project to access
#d_arg job_list List of job ids
#d_arg db_display A list of the database parameters to be displayed
#d_arg db_display_format A list of the field widths (as integers) to \
use for displaying the items in the db_display list.
  if { [regexp ^OK$ [DbHandlerRequest desc_list \
	  "DbReturnRecord" $project -list $job_list -list $db_display \
			 -list $db_display_format]] } {
      return $desc_list
  }
  Report 1 "ERROR Extracting data: $desc_list"
  return ""
}

#-----------------------------------------------------------------------
proc DbStartHandler { { broadcastCmd {} } } {
#-----------------------------------------------------------------------
#d_sum Start a database handler, or connect to one already running
#d_desc This command opens a server socket and starts a new handler \
process in background. The handler process sends back a port number \
for the client application to use when connecting to the handler.
#d_desc Optionally the calling process can specify a callback \
procedure which will be invoked each time a broadcast message is \
received from the handler. The callback function must be defined in \
the application.
#d_arg broadcastCmd Procedure to be invoked for broadcast messages

  global handler
  # Register the callback function
  set handler(DB_BROADCAST_CALLBACK) $broadcastCmd
  set status 0

  # Check whether a handler is already running
  set port [GetHandlerPort -nowait]
  if { $port > -1 } {
      Report 1 "DbStartHandler: handler already running using port number $port"
      set handler(DB_PORT) $port
      set status 1
  } else {
      # Start the handler in background
      if { ![ catch { eval exec $handler(RUN_DBSERVER) & } ] } {
	  # Wait for the port number to be updated
	  Report 1 "DbStartHandler: waiting for update"
	  set port [GetHandlerPort]
	  Report 1 "DbStartHandler: update received"
	  if { $port > -1 } {
	      Report 1 "DbStartHandler: handler using port number $port"
	      set handler(DB_PORT) $port
	      set status 1
	  } else {
	      Report 1 "DbStartHandler: Bad port number returned"
	  }
      }  
  }
  set handler(DB_HANDLER_STATUS) $status
  Report 1 "DbStartHandler: status = $handler(DB_HANDLER_STATUS)"
  return $status
}


#-----------------------------------------------------------------------
proc OpenDBsql { dbname dir } {
#-----------------------------------------------------------------------
#d_sum Set the values of one or more data items
#d_desc This sets the values of multiple data items for a particular job \
in the specified project database. The argument list should consist of \
key-value pairs.
#d_desc Setting multiple data items in a single SetData command is \
recommended as this minimises the number of socket transactions required \
to send the data to the serve
#d_arg job_id  Number of the job for which the data is being s
# Use eval to invoke the command, to stops the arglist concatenated
# into a single argument
  if { ![regexp ^OK$ \
	  [eval DbHandlerRequest message "OpenDB_sql" \
	       $dbname $dir]] } {
      Report 1 "ERROR Updating database: $message"
      return 0
  }
  return 1
}


#-----------------------------------------------------------------------
proc SetDatasql { job_id args } {
#-----------------------------------------------------------------------
#d_sum Set the values of one or more data items
#d_desc This sets the values of multiple data items for a particular job \
in the specified project database. The argument list should consist of \
key-value pairs.
#d_desc Setting multiple data items in a single SetData command is \
recommended as this minimises the number of socket transactions required \
to send the data to the serve
#d_arg job_id  Number of the job for which the data is being s
# Use eval to invoke the command, to stops the arglist concatenated
# into a single argument
  if { ![regexp ^OK$ \
	  [eval DbHandlerRequest message "SetData_sql" \
	       $job_id $args]] } {
      Report 1 "ERROR Updating database: $message"
      return 0
  }
  return 1
}

#-----------------------------------------------------------------------
proc NewProjectsql { projectname } {
#-----------------------------------------------------------------------
#d_sum Set the values of one or more data items
#d_desc This sets the values of multiple data items for a particular job \
in the specified project database. The argument list should consist of \
key-value pairs.
#d_desc Setting multiple data items in a single SetData command is \
recommended as this minimises the number of socket transactions required \
to send the data to the server.
#d_arg dbname Alias of the project database
#d_arg dir  Number of the job for which the data is being set
  # Use eval to invoke the command, to stops the arglist concatenated
  # into a single argument
  if { ![regexp ^OK$ \
	  [eval DbHandlerRequest pid "NewProject_sql" \
	       $projectname]] } {
      Report 1 "ERROR Updating database: $pid"
      return 0
  }
  return $pid
}

#-----------------------------------------------------------------------
proc NewJobsql { projectname } {
#-----------------------------------------------------------------------
#d_sum Set the values of one or more data items
#d_desc This sets the values of multiple data items for a particular job \
in the specified project database. The argument list should consist of \
key-value pairs.
#d_desc Setting multiple data items in a single SetData command is \
recommended as this minimises the number of socket transactions required \
to send the data to the server.
#d_arg dbname Alias of the project database
#d_arg dir  Number of the job for which the data is being set
  # Use eval to invoke the command, to stops the arglist concatenated
  # into a single argument
  if { ![regexp ^OK$ \
	  [eval DbHandlerRequest jobid "NewJob_sql" \
	       $projectname]] } {
      Report 1 "ERROR Updating database: $jobid"
      return 0
  }
  return $jobid
}

#-----------------------------------------------------------------------
proc GetDatasql { jobid args } {
#-----------------------------------------------------------------------
#d_sum Set the values of one or more data items
#d_desc This sets the values of multiple data items for a particular job \
in the specified project database. The argument list should consist of \
key-value pairs.
#d_desc Setting multiple data items in a single SetData command is \
recommended as this minimises the number of socket transactions required \
to send the data to the server.
#d_arg dbname Alias of the project database
#d_arg dir  Number of the job for which the data is being set
  # Use eval to invoke the command, to stops the arglist concatenated
  # into a single argument
  if { ![regexp ^OK$ \
	  [eval DbHandlerRequest value "GetData_sql" \
	       $jobid $args]] } {
      Report 1 "ERROR Updating database: $value"
      return 0
  }
  return $value
}

#-----------------------------------------------------------------------
proc CloseDBsql { } {
#-----------------------------------------------------------------------
#d_sum Set the values of one or more data items
#d_desc This sets the values of multiple data items for a particular job \
in the specified project database. The argument list should consist of \
key-value pairs.
#d_desc Setting multiple data items in a single SetData command is \
recommended as this minimises the number of socket transactions required \
to send the data to the server.
#d_arg dbname Alias of the project database
#d_arg dir  Number of the job for which the data is being set
  # Use eval to invoke the command, to stops the arglist concatenated
  # into a single argument
  if { ![regexp ^OK$ \
	  [eval DbHandlerRequest message "CloseDB_sql" \
	       ]] } {
      Report 1 "ERROR Updating database: $message"
      return 0
  }
  return 1
}

# Commands for talking to the database handler

#-----------------------------------------------------------------------
proc DbHandlerConnect { } {
#-----------------------------------------------------------------------
#d_sum Open socket and connect to (register with) database handler
#d_desc This command should only be invoked by the client application \
after DbStartHandler, which establishes the correct port number to use \
to communicate with the database server process.
#d_desc It sets up a client socket with a callback procedure \
(processHandlerMessage) which is invoked whenever the socket becomes \
readable (corresponding to data being sent from the handler). This \
command then connects to the handler, allowing the other database \
requests to be made via the appropriate API commands.
#d_desc Return 1 on success and 0 on failure
  global handler

  # FIXME Trap for the client invoking this more than once

  Report 1 "DbHandlerConnect: connection request"
  set sock [socket localhost $handler(DB_PORT)]
  set handler(DB_SOCKET) $sock

  # Set the socket to operate asynchronously (i.e. non-blocking)
  fconfigure $sock -buffering line -blocking 0
  fileevent $sock readable "processHandlerMessage $sock"

  # Initialise the handler array for dealing with sending
  # requests and receiving responses
  set handler(REQUEST_NUMBER) 0
  set handler(REQUEST_LIST) {}
  set handler(ERROR_MESSAGE) {}

  # Register with handler
  set handler(DB_CONNECTION_STATUS) 1
  set status [DbHandlerRequest message "DbRegister" [GetUserId] [pid]]
  Report 1 "DbHandlerConnect: got back status \"$status\" message \"$message\""

  if { [regexp "^failed\$" $status] } {
      set handler(DB_CONNECTION_STATUS) 0
      return 0
  } else {
      
  }
  return $handler(DB_CONNECTION_STATUS)
}

#-----------------------------------------------------------------------
proc DbHandlerDisconnect { } {
#-----------------------------------------------------------------------
#d_sum Disconnect socket from the database handler
  global handler
  Report 1 "DbHandlerDisconnect:  disconnection request"
  if { ! $handler(DB_CONNECTION_STATUS) } {
    Report 1 "DbHandlerDisconnect: currently no connection to handler"
    return 0
  } else {
    DbHandlerRequest message "DbDisconnect"
    Report 1 "DbHandlerDisconnect: got back $message"
    close_db_connection
    return 1
  }
}


#-----------------------------------------------------------------------
proc close_db_connection { } {
#-----------------------------------------------------------------------
#d_sum Internal procedure
#d_desc Remove socket connection and reset internal parameters
  global handler
  if { $handler(DB_CONNECTION_STATUS) } {
    catch { close $handler(DB_SOCKET) }
  }
  set handler(DB_SOCKET) ""
  set handler(DB_CONNECTION_STATUS) 0
  return
}

#-----------------------------------------------------------------------
proc DbShutdown { } {
#-----------------------------------------------------------------------
#d_sum Issue shutdown request to the handler
   global handler
   Report 1 "DbShutdown: shutdown request"
   if { ! $handler(DB_CONNECTION_STATUS) } {
       Report 1 "DbShutdown: currently no connection to handler"
       return 0
   } else {
       DbHandlerRequest message "ShutDown"
       Report 1 "DbShutdown: got back $message"
       close_db_connection
       set handler(DB_HANDLER_STATUS) 0
       return 1
   }
}

#-----------------------------------------------------------------------
proc DbHandlerRequest { messageVar command args } {
#-----------------------------------------------------------------------
#d_sum Send a request to the database handler and return the response.
#d_desc The request consists of the all arguments in args, excluding the \
After the request is sent to the handler this procedure waits \
until a response is recieved i.e. it operates in a pseudo-synchronous \
manner.
#d_desc Responses from the handler consist of Tcl lists - the first \
element in the list is the return value of this procedure, the \
remainder is returned in the variable named by the last argument in \
args.
#d_desc In case of failure, ERROR will be returned.
#d_arg messageVar Name of variable used to return output to the \
calling procedure
#d_arg command Command to be sent to the handler

  global handler

  # Echo input
  Report 1 "DbhandlerRequest:"
  Report 1 "messageVar: $messageVar"
  Report 1 "command   : $command"
  Report 1 "args      : $args"

  # Determine the variable to send any messages back in
  upvar 1 $messageVar message
  set message ""

  # Check for connection
  if { ! $handler(DB_CONNECTION_STATUS) } {
    Report 1 "DbHandlerRequest: currently no connection to handler"
    return 0
  }

  # Set up an id for the request
  # This will be sent back by the handler as part of a response
  set id ""
  incr handler(REQUEST_NUMBER)
  append id "request\#" "$handler(REQUEST_NUMBER)"
  Report 1 "Request   : $handler(REQUEST_NUMBER)"
  # Add to the list of requests
  lappend handler(REQUEST_LIST) $id
  Report 1 "DbHandlerRequest: current request list: $handler(REQUEST_LIST)"

  # Build the request and wrap it in pseudo-XML
  set request [eval WrapDbRequest $id $command $args]
  Report 1 "DbHandlerRequest: request is \"$request\""
  # Store the command for reporting elsewhere
  set handler(REQUEST_CMD) $request

  # Check we have a connection to the handler
  if { ![info exists handler(DB_SOCKET)] } {
    ##puts "DbHandlerRequest: no connection to handler"
    set message "No connection to handler"
    return "ERROR"
  }
  set client $handler(DB_SOCKET)
  set timeout $handler(DB_TIMEOUT)

  set line ""

  # Send the request to the server
  ##puts "DbHandlerRequest: request \"$request\""
  if { ![catch { puts $client $request } msg] } {
    Report 1 "DbHandlerRequest: request sent ok"
    # Wait for response
    # This is done by waiting for an update on handler(REQUEST_STATUS)
    # We also set up a timeout to prevent waiting forever e.g. if
    # there is some problem
    set handler(REQUEST_ID) $id
    set handler(REQUEST_STATUS_$id) 1
    after $timeout set handler(REQUEST_STATUS_$id) "TIMEOUT"
    Report 1 "DbHandlerRequest: waiting for response ..."
    vwait handler(REQUEST_STATUS_$id)
    after cancel set handler(REQUEST_STATUS_$id) "TIMEOUT"
    if { $handler(REQUEST_STATUS_$id) == "TIMEOUT" } {
       Report 1 "DbHandlerRequest: ... timed out waiting for response to $id"
       set message "Handler response timed out waiting for $id"
       set status ERROR
    } else {
       Report 1 "DbHandlerRequest: ... response received for $id"
       set message $handler(REQUEST_RESPONSE_$id)
       set status $handler(REQUEST_RESPONSE_STATUS_$id)
       Report 1 "DbHandlerRequest: got back status \"$status\" \"$message\""
    }
    return $status
  } else {
    Report 1 "DbHandlerRequest: failed to send request"
    set message "Failed to send request: $msg"
    return "ERROR" 
  }
}

#-----------------------------------------------------------------------
proc DbRequestsPending { } {
#-----------------------------------------------------------------------
#d_sum Returns the number of pending requests within the client API
#d_desc This procedure allows the wait status of the API to be \
determined by the calling application.
#d_desc It returns the number of requests queued up and waiting for \
responses from the database server.
  global handler
  return [llength $handler(REQUEST_LIST)] 
}

#-----------------------------------------------------------------------
proc processHandlerMessage { client } {
#-----------------------------------------------------------------------
#d_sum Handler for messages recieved from the database server.
#d_desc This processes messages recieved from the handler process via \
the socket connection. Messages should fall into two catergories, \
either: 1. responses to requests sent via DbHandlerRequest, or \
2. unsolicited "broadcast" messages from the handler.
#d_arg client Socket id
    # Handle messages from the server back to the client
    global handler

    Report 1 "processHandlerMessage: invoked with client $client"
    Report 1 "*** REQUEST_LIST: $handler(REQUEST_LIST)"

    # Called whenever the socket is readable
    set line ""
    if { ![catch { set count [gets $client line] } msg] } {
	Report 1 "processHandlerMessage: from server: \"$line\" msg: $msg"

	# Check for count from gets less than zero
        # If varName is specified and an empty string is returned in
        # varName  because of end-of-file or because of insufficient
        # data in nonblocking mode, then the  return  count  is  -1.
	# Note also that at EOF gets returns a blank string but doesn't
	# consume any input, so it is possible to end up in an
	# infinite loop in these situations
	if { $count < 0 } {
	    # Check for eof
	    if { [eof $client] } {
		Report 1 "processHandlerMessage: end of file"
		Report 1 "*** Connection to the server has been lost ***"
		# FIXME: this should invoke some catastrophe handling
		# procedure
		close_db_connection
	    }
	    Report 1 "processHandlerMessage: count < 0, discarding"
	    return
	}

	# Are we waiting for a response?
        if { [llength $handler(REQUEST_LIST)] > 0 } {
	    # Attempt to unwrap the response and check if the ids match
	    if { [UnwrapDbResponse $line status result response_id] } {
		# We are expecting a message from the server in
		# response to a request
		Report 1 "processHandlerMessage: id of response = $response_id"
		if { [set k [lsearch $handler(REQUEST_LIST) $response_id]] > -1 } {
		    # The id is in the list
		    set handler(REQUEST_RESPONSE_$response_id) $result
		    set handler(REQUEST_RESPONSE_STATUS_$response_id) $status
		    set handler(REQUEST_STATUS_$response_id) 0
		    # Remove the request from the list
		    set handler(REQUEST_LIST) [lreplace $handler(REQUEST_LIST) $k $k]
		    return
		} else {
		    # Id not in the list
		    Report 1 "processHandlerMessage: response id not in list"
		}
	    } else {
		Report 1 "processHandlerMessage: this doesn't appear to be a response"
	    }
	} else {
	    Report 1 "processHandlerMessage: no requests pending"
	}

	# Test for broadcast message
	if { [UnwrapDbBroadcast $line message] } {
	    Report 1 "processHandlerMessage: broadcast message from the handler"
	    if { $handler(DB_BROADCAST_CALLBACK) != "" } {
		# Issue callback
		Report 1 "processHandlerMessage: invoking callback \"$handler(DB_BROADCAST_CALLBACK)\""
		if { [catch { $handler(DB_BROADCAST_CALLBACK) "$message" } errmsg] } {
		    # Callback failed
		    Report 1 "processHandlerMessage: callback failed: $errmsg"
		}
		# Drop out now
		return
	    }
	} else {
	    Report 1 "processHandlerMessage: not a broadcast message"
	}

	# At this point we don't know what the message is
	Report 1 "processHandlerMessage: unrecognised message"
	return

    } else {
	# The initial read caught an error
	Report 1 "processHandlerMessage: error: $msg"
        set handler(ERROR_MESSAGE) $msg
	return
    }
}

#-----------------------------------------------------------------------
proc DbVerifyConnection { } {
#-----------------------------------------------------------------------
#d_sum Check whether the connection to the handler is active
#d_desc Return 1 if the connection exists and is working, return 0 \
otherwise

}

#-----------------------------------------------------------------------
proc DbClientError { } {
#-----------------------------------------------------------------------
#d_sum Return the last error message from inside the client API
#d_desc This returns the error message (blank if no error occurred since \
last time the error was read) and resets the message.
  global handler
  set msg $handler(ERROR_MESSAGE)
  set handler(ERROR_MESSAGE) {}
  return $msg
}

#-----------------------------------------------------------------------
proc GetHandlerPort { args } {
#-----------------------------------------------------------------------
# This is a fast and dirty way to acquire the port number for the
# server process.
    # The lockfile is called dbccp4i.LOCK and has the following content:
# port number is:4090
    set port -1
    set thistry 0
    set ntries 10
    set wait_time 1000

    if { [lsearch $args "-nowait" ] > -1 } {
	set ntries 1
	set wait_time 0
    }

    set lockfile [file join [GetDotCCP4] dbccp4i.LOCK]
    while { $thistry < $ntries } {
	if { [file exists $lockfile] } {
	    # Open and read the lockfile contents
	    Report 1 "GetHandlerPort: opening lockfile"
	    set fp [open $lockfile "r"]
	    if { [gets $fp line] > 0 } {
		Report 1 "GetHandlerPort: line from lockfile is $line"
		set port [lindex [split $line ":"] end]
	    }
	    close $fp
	    Report 1 "GetHandlerPort: port $port"
	    # Test the port
	    # This is a little messy but does the job for now
	    if { ![catch { set sock [socket localhost $port] } err] } {
		# Connected - issue status request
		fconfigure $sock -buffering line
		set request [eval WrapDbRequest 0 "DbRequestStatus"]
		Report 1 "Request is: $request"
		puts $sock $request
		gets $sock line
		Report  1 "Response is $line"
		if { [UnwrapDbResponse $line status result response_id] } {
		    if { [regexp "^OK$" $status] } {
			# Verified that the server is active
			# Close socket and return port number
			Report 1 "Socket opened and server is active - ok"
			## Before close socket, disconnect handler
			set request [eval WrapDbRequest 0 "DbDisconnect"]
			Report 1 "Request is: $request"
			puts $sock $request
			close $sock
			return $port
		     }
		}
		# No verification
		Report 1 "Socket opened but server is inactive?"
		close $sock
	    }
	    # Failed to open socket
	    Report 1 "GetHandlerPort: failed to open socket, retry \#$thistry..."
	}
	# Wait for a while then retry
	after $wait_time
	incr thistry
    }
    Report 1 "GetHandlerPort: timed out waiting for lockfile"
    return -1
}

#-----------------------------------------------------------------------
proc CheckServerStatus { } {
#-----------------------------------------------------------------------
    # Check if the handler is running
    # Returns 1 if the server is connected and responds to the status
    # request, 0 otherwise   
    if { [regexp "^OK$" [DbHandlerRequest msg "DbRequestStatus"]] } {
	return 1
    } else {
	return 0
    }
}

#=======================================================================
# PLACEHOLDER FUNCTIONS
#=======================================================================
# A set of functions that replicate or replace that found in CCP4i
# and enable this code to be used standalone
# Possibly should be migrated to a systemcore.tcl and shared with
# main CCP4i?
# Also compare the duplicated functions in the ccp4i.py Python module.

#PJX duplicates function from system.tcl
#PJX note that this is missing GetTmpFileName, DeleteFile, BinPath,
#PJX and Execute - so it will fail under WINDOWS.
#-----------------------------------------------------------------------
proc GetUserId { } {
#-----------------------------------------------------------------------
#d_sum Return the user's id.
  if { ![regexp WINDOWS [GetOPSYS]] } {
      return [GetEnvPath USER]
  } else {
    set log [GetTmpFileName]
    if { [Execute [BinPath iam] {} status report -log $log] &&
    		[ReadFile $log logtext] } {
      DeleteFile $log
      return [string trim $logtext]
    } else {
      return 'UNKNOWN'
    }
  }
}

#-----------------------------------------------------------------------
proc Report { code message } {
#-----------------------------------------------------------------------
    # Dummy function to report an error or message
    set logfile [open [file join [GetDotCCP4] dbclientapi.log] "a+"]
    puts $logfile $message
    close $logfile
    #puts $message
    return
}

#PJX This code taken from ccp4i/src/system.tcl
#-----------------------------------------------------------------------
proc GetDotCCP4 { } {
#-----------------------------------------------------------------------
#d_sum Return the location of the .CCP4 directory
#d_desc The .CCP4 directory is used to store user-specific CCP4 data.

  switch [GetOPSYS] \
  UNIX {
    return [file join [GetEnvPath HOME] .CCP4]
  }
  WINDOWS  {
    if { [GetEnvPath USERPROFILE ] == "" } {
      return [file join [GetEnvPath SystemRoot] Profiles $system(LOGIN_NAME) CCP4]
    } else { 
      return [file join [GetEnvPath USERPROFILE] CCP4]
    }
  }
  return {}
}

#PJX This procedure should probably be migrated into CCP4i.
#-----------------------------------------------------------------------
proc GetOPSYS { } {
#-----------------------------------------------------------------------
#d_sum Return the operating system type
#d_desc Returns the upper-cased platform type e.g. UNIX or WINDOWS.
    global system tcl_platform
    if { [array exists system] && [info exists system(OPSYS)] } {
	return $system(OPSYS)
    } else {
	return [string toupper $tcl_platform(platform)]
    }
}

#PJX This procedure taken from ccp4i/src/system.tcl and modified
#PJX to use GetOPSYS procedure
#----------------------------------------------------------------
proc GetEnvPath { var { report 1 } } {
#----------------------------------------------------------------
#d_sum Get operating system environment variable
#d_desc Return a path name associated with an environment variable - in a \
Windows system the separators are converted to forward slash. \
Return an empty string if no environment variable found.
#d_arg var Environment variable
#d_arg report Optional - default 1.  If true (1) then output warning message if \
environment variable not found.
  global env

# Get environment variable - cope with whether input var has
# leading '$' or not

  regsub {^\$} $var  {} var1

# On NT change the separators to unix style /

  if { [regexp WINDOWS [GetOPSYS]] } {
    set status [catch { global env; 
      regsub -all {\\} $env($var1) \/ path } nc ]
  } else {
    set status [catch {global env; set p1 $env($var1)} path ]
  }
  if { $status } { 
    set path ""
    if { $report } { 
      WarningMessage "Cannot get value for environment variable \"$var\"" }
  } 
  return $path
}

#PJX Non-graphical placeholder for WarningMessage function in CCP4i
proc WarningMessage { message } {
    Report 0 $message
}

#-----------------------------------------------------------------------
proc DbGetTimeout { } {
#-----------------------------------------------------------------------
#d_sum Returns the timeout used in DbHandlerRequest
#d_desc The timeout is the length of time that the client API will wait \
after sending a request to the handler, before giving up and returning \
control to the client application.
#d_desc The timeout is a number of milliseconds, with a minimum value \
of zero.
    global handler
    return $handler(DB_TIMEOUT)
}

#-----------------------------------------------------------------------
proc DbSetTimeout { timeout } {
#-----------------------------------------------------------------------
#d_sum Set the timeout used in DbHandlerRequest
#d_arg timeout Length of the timeout in milliseconds
    global handler
    if { $timeout < 0 } {
	set handler(DB_TIMEOUT) 0
    } else {
	set handler(DB_TIMEOUT) $timeout
    }
    return $handler(DB_TIMEOUT)
}

#-----------------------------------------------------------------------
proc DbSupported { DB} {
#-----------------------------------------------------------------------
#d_sum Set the values of one or more data items
#d_desc This sets the values of multiple data items for a particular job \
in the specified project database. The argument list should consist of \
key-value pairs.
#d_desc Setting multiple data items in a single SetData command is \
recommended as this minimises the number of socket transactions required \
to send the data to the serve
#d_arg job_id  Number of the job for which the data is being s
# Use eval to invoke the command, to stops the arglist concatenated
# into a single argument
  if { ![regexp ^OK$ \
	  [eval DbHandlerRequest message "DbSupported" \
	       $DB]] } {
      Report 1 "ERROR Updating database: $message"
      return 0
  }
  return $message
}





