#
#     Copyright (C) 2006
#
#     This code is distributed under the terms and conditions of the
#     CCP4 Program Suite Licence Agreement as a CCP4 Library.
#     A copy of the CCP4 licence can be obtained by writing to the
#     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
#
#====================================================================
# CCP4 Interface - dbsocket.py
#
# Classes and methods for handling pseudo-asynchronous communications
# with the db handler
# 
# Peter Briggs
#
#====================================================================

#CCP4i_cvs_Id $Id: dbsocket.py,v 1.8 2007/01/23 11:58:40 wy45 Exp $

##################################################################

import socket
import select
import time
import threading
import Queue
import re
import exceptions
import xml.dom.minidom
import xml.sax.saxutils as su

# Module globals
_request_id = 0
_debug = False
# The idea is that the client API should use this as the way to
# communicate with the handler.
# The client would need to implement its own listener to
# check for broadcast messages - this would be important for a
# visualiser application which would need to respond to changes
# in the db content

class dbRequest:
    """Class representing a handler request."""

    def __init__(self,command,arguments):
        """Command is the base command being issued; arguments is
        a tuple of appropriate options"""
        self.__command = command
        self.__arg_list = arguments
        self.__id = get_request_id()

    def __str__(self):
        """Print the string representation of the dbRequest object."""
        text = str(self.__command)
        if len(self.__arg_list) !=0:
            for arg in self.__arg_list:
                text = text + " " + str(arg)
        return text

    def xml(self):
        """Transform the request into XML"""
        xml_request = "<db_request><command>"+su.escape(str(self.__command))+"</command>"
        if len(self.__arg_list) !=0:
            for arg in self.__arg_list:
                if type(arg) == list:
                    xml_request = xml_request + "<argument><list>"
                    for item in arg:
                        xml_request = xml_request + "<item>"+ su.escape(str(item))+"</item>"
                    xml_request = xml_request + "</list></argument>"
                else:
                    xml_request = xml_request + \
                              "<argument>"+su.escape(str(arg))+"</argument>"
        xml_request = xml_request + \
                      "<request_id>"+self.__id+"</request_id></db_request>"
        return xml_request

    def request_id(self):
        """Return the request id string."""
        return self.__id

class dbResponse:
    """Class representing a response from the handler.

    The object is initialised with the raw XML returned by the
    handler, and is classified as either a response to request
    or as a broadcast."""

    def __init__(self,xml_response):
        """xml_response is the raw response from the handler"""
        self.__xml_response = xml_response
        # Parameters for response
        self.__is_response = False
        self.__status = ""
        self.__result = ""
        self.__id = ""
        # Parameter for broadcast
        self.__is_broadcast = False
        self.__broadcast = ""
        # Parse the input
        self.parse()

    def __str__(self):
        """String representation of the response object."""
        if self.is_broadcast():
            return "Broadcast: "+str(self.broadcast())
        elif self.is_response():
            return "Response: "+str(self.status())+" "+str(self.result())
        else:
            return "Unrecognised: "+str(self.__xml_response)

    def parse(self):
        """Attempt to process the XML and extract the components.

        Return False if the extraction fails, and True otherwise."""
        
        # Check for empty string
        if self.__xml_response == "":
            ##print "Empty string in dbResponse parse method"
            return

        # Test for broadcast first
        rbroadcast = re.compile("<db_broadcast><message>([^<>]*)</message></db_broadcast>")
        try:
            broadcast = rbroadcast.match(self.__xml_response)
            if broadcast:
                self.__broadcast = broadcast.group(1)
                self.__is_broadcast = True
                return
        except exceptions.Exception,e:
            # Parsing failed
            print "Exception in dbResponse parse method"
            print "Input xml: "+str(self.__xml_response)
            print "response.parse (broadcast): Exception: "+str(e)
            raise
            
        try:
            doc = xml.dom.minidom.parseString(self.__xml_response)
            # Status tag
            statusTag = doc.getElementsByTagName("status")
            status = statusTag[0].childNodes[0].data
            # Result tag
            resultTag = doc.getElementsByTagName("result")
            # Id tag
            # if single value, append to resultlist, else retrive
            # items from nested tag
            if resultTag[0].childNodes[0].nodeType == \
                   resultTag[0].childNodes[0].TEXT_NODE:
                result = resultTag[0].childNodes[0].data   
            else:
                list = resultTag[0].childNodes
                items = list[0].childNodes
                result = self.getXMLitems(items,[])
            try:
                responseidTag = doc.getElementsByTagName("response_id")
                responseid = responseidTag[0].childNodes[0].data
            except IndexError:
                # The response id is not compulsory
                responseid = ""

            # Successfully processed the XML
            self.__status = status
            self.__result = result
            self.__id = responseid
            self.__is_response = True
            return
        
        except IndexError:
            # Parsing failed
            print "response.parse (response): not a response:"
            print str(self.__xml_response)
            return

    def getXMLitems(self,items,varlist):
        """Retrieve items from an XML-wrapped result list.

        This is called from the parse method to deal with
        lists in the XML response i.e.
        <list><item>..</item>...</list> constructs."""
            
        value = varlist
        tmp = []
        for item in items:
            # check if it is nested
            try:
                if item.childNodes[0].nodeName == 'list':
                    # if find the next node is 'list', get the nested items
                    nesteditems = item.childNodes[0].childNodes
                    # call the function itself
                    self.getXMLitems(nesteditems,tmp)
                else:
                    # if it is a text, get the data from the node
                    tmp.append(item.childNodes[0].data)
            except IndexError:
                # This seems to happen when the "item" node contains no
                # data - append an empty data item
                tmp.append(u"")
        # append the tmp result to the return value
        value.append(tmp)
    
        if len(value) == 1:
            # only have one element in the list, then return the element. 
            return value[0]
        else:
            return value

    def is_response(self):
        """Check if the object contains an XML response."""
        return self.__is_response

    def status(self):
        """Return the status component of the response."""
        if self.__is_response:
            return self.__status
        else:
            return ""

    def result(self):
        """Return the result component of the response."""
        if self.__is_response:
            return self.__result
        else:
            return ""

    def response_id(self):
        """Return the id component of the response."""
        if self.__is_response:
            return self.__id
        else:
            return ""

    def is_broadcast(self):
        """Check if the object contains an XML broadcast."""
        return self.__is_broadcast

    def broadcast(self):
        """Return the broadcast message from the XML."""
        if self.__is_broadcast:
            return self.__broadcast
        else:
            return ""

    def report(self):
        """Report the contents of the response object."""
        if self.is_broadcast():
            print "Broadcast message:"
            print "\t\""+str(self.broadcast())+"\""
        elif self.is_response():
            print "Request response:"
            print "\tStatus: "+str(self.status())
            print "\tResult: "+str(self.result())
            print "\tId    : "+str(self.response_id())
        else:
            print "Unrecognised:"
            print "\t"+str(self.__xml_response)

class dbsocket(threading.Thread):
    """Class for handling interactions between client and handler.

    A dbsocket object is instantiated by providing it with a host name
    and port number. It then opens a socket connection to the handler
    via that host port. The client can send requests to the handler,
    (and get back the responses) using the 'request' method of the
    dbsocket object.

    The dbsocket intercepts handler broadcasts and ensures that these
    don't interfere with the client's request/response pairs. The
    client can check and collect broadcasts separately using the
    nbroadcast and getbroadcast methods.

    Responses to requests are in the form of lists: [status,result]."""
    
    def __init__(self,host,port,timeout=5):
        
        # Start
        ##print "Instantiating dbsocket object..."
        threading.Thread.__init__(self)
        self.requestQueue = Queue.Queue()
        self.responseQueue = Queue.Queue()
        self.broadcastQueue = Queue.Queue()
        # Setup the socket connection
        try:
            self.host = host
            self.port = port
            self.addr = (self.host,self.port)
            self.socket = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
            self.socket.connect(self.addr)
        except exceptions.Exception,e:
            # Connection attempt failed
            print "dbsocket: init failed with exception:"
            print str(e)
            raise
        # Start the queue
        self.active = True
        self.timeout = timeout
        self.request_id = 0
        print "dbsocket: invoking start method"
        self.start()

    # Check that the dbsocket is still active
    def isActive(self):
        """Return True if the dbsocket is still active, False if not."""
        return self.active

    # Application should use this to send a
    # request to the handler
    def request(self,command,args):
        """Send a request to the handler.

        The request consists of a command and a list of arguments. The
        result of the request method is the response from the handler
        (or a timeout message if no response was received within the
        timeout period)."""

        global _debug
        debug = _debug
        
        request = dbRequest(command,args)
        if debug: print "dbsocket: received request \""+str(request)+"\""
        self.requestQueue.put(request)
        if debug: print "dbsocket: request placed in the queue, waiting"
        response = self.responseQueue.get()
        if debug: print "dbsocket: got the response \""+str(response)+"\""
        if debug: print "dbsocket: returning"
        return response

    # Start the dbsocket process running
    def run(self):
        """Called by start method to start the dbsocket process.

        This handles sending requests to the handler, processing
        the responses, and dealing with any broadcast messages."""
        
        global _debug
        debug = _debug
        
        # Use start_time as a flag to indicate whether we're
        # currently waiting for a request
        start_time = -1
        # Set up a loop to run whileever the dbsocket is active
        xml_response = ""
        while self.active:
            ##if debug: print "In loop..."
            if self.requestQueue.qsize() > 0 and start_time < 0:
                # Get the next request and send it
                request = self.requestQueue.get()
                xml_request = request.xml()
                if debug: print "dbsocket.run: got request: \""+str(xml_request)+"\""
                self.socket.send(str(xml_request)+"\n")
                if debug: print "dbsocket.run: request sent"
                # Get the current time in case of timeout
                start_time = time.time()
                if debug: print "dbsocket.run: start_time = "+str(start_time)
            if socket_is_readable(self.socket):
                if debug: print "dbsocket.run: socket is readable"
                # There is some data to read
                xml_response = xml_response+str(self.socket.recv(4096))
                if debug: print "dbsocket.run: read data from socket: \""+str(xml_response)+"\""
                
                # Process the response
                if xml_response != "" and xml_response[-1] == "\n":
                    response = dbResponse(xml_response)
                    if xml_response == "":
                        # Empty string received from the handler
                        if debug: print "dbsocket.run: empty string received from handler"
                        self.responseQueue.put(["ERROR","Empty string received"])
                        # Clear the current request
                        start_time = -1
                    elif response.is_broadcast():
                        # Add to the broadcast queue
                        if debug: print "dbsocket.run: detected broadcast"
                        print str(response.broadcast())
                        self.broadcastQueue.put(response.broadcast())
                    elif start_time >= 0.0:
                        # Check whether this is a response to a request
                        if response.is_response():
                            if response.response_id() == request.request_id():
                                # This is the correct response
                                # Add it to the response queue
                                if debug: print "dbsocket.run: response id matched request id"
                                self.responseQueue.put([response.status(),response.result()])
                                # Reset the timer
                                start_time = -1
                            else:
                                # A response but with the wrong id
                                if debug:
                                    print "dbsocket.run: bad response received: \""\
                                          +str(xml_response)+"\" (ignored)"
                        else:
                            # Not sure what this is
                            if debug: print "dbsocket.run: error assumed"
                            report_error(response.result())
                    # Reset the response buffer
                    xml_response = ""
            ##if debug: print "dbsocket.run: checking for timeout..."
            if self.timeout_exceeded(start_time):
                # No response from server after timeout
                # has elapsed
                if debug: print "dbsocket.run: timeout condition reached"
                self.responseQueue.put(["ERROR","Server timed out"])
                # Clear this request
                start_time = -1

    # Close the connection
    def close(self):
        """Close the dbsocket connection.

        Terminate the socket connection to the handler and flag
        the dbsocket as inactive."""
        
        self.active = False
        self.socket.close
        return True	

    # Return true if timeout period has been exceeded
    def timeout_exceeded(self,start_time):
        """Internal method: check whether the timeout has been exceeded.

        Returns True if the elapsed time has exceeded the timeout
        period, and False if not."""
        
        if start_time < 0:
            return False
        if (time.time() - start_time) > self.timeout:
            return True
        return False

    # Return number of broadcast messages in the queue
    def nbroadcasts(self):
        """Return the number of broadcast messages waiting processing."""
        return self.broadcastQueue.qsize()
    
    # Pull the oldest broadcast message from the queue
    def getbroadcast(self):
        """Return the oldest available broadcast message for processing.

        Broadcast messages are put in a queue when received. Fetching
        a broadcast message removes it from the queue."""
        
        if self.nbroadcasts() > 0:
            return self.broadcastQueue.get()
        else:
            return ""

# Check and report broadcasts
class report_broadcast(threading.Thread):
    """Example of a thread process to monitor and report on broadcasts.

    Given a dbsocket object, the report_broadcast object continually
    checks for broadcast messages and prints them to stdout."""
    
    def __init__(self,dbsock):
        threading.Thread.__init__(self)
        self.setDaemon(1)
        self.socket = dbsock
        self.start()

    def run(self):
        while self.socket.isActive():
            while self.socket.nbroadcasts() > 0:
                broadcast = self.socket.getbroadcast()
                print "******************************************************"
                print "Received broadcast message:\n"
                print str(broadcast)
                print "******************************************************"

# Test whether a specific socket is readable
def socket_is_readable(socket):
    """Checks whether the specified socket object is readable.

    Returns True if data can be read from the socket, False if not."""
    
    if not socket:
        return False
    try:
        slist = select.select([socket],[],[],0.01)
        if len(slist[0]) > 0:
            # Socket is ready to read
            return True
    except exceptions.Exception,e:
        print "Exception in socket_is_readable: "+str(e)
    return False

# Generate a unique request id
def get_request_id():
    """Generate a unique identifier string for a request."""
    global _request_id
    _request_id = _request_id + 1
    requestid = "request#"+str(_request_id)
    return requestid

# Report error
def report_error(error):
    """Internal error reporting method."""
    
    print "dbsocket: error \""+str(error)+"\""

# Test/example

if __name__ == "__main__":
    # Example of request generation
    myRequest = dbRequest("DbCommand",["arg1","arg2","arg3"])
    print str(myRequest)
    print str(myRequest.xml())
    myRequest = dbRequest("DbCommand2",["arg4"])
    print str(myRequest)
    print str(myRequest.xml())
    myRequest = dbRequest("DbCommand3",[])
    print str(myRequest)
    print str(myRequest.xml())
    # Example of response unwrapping
    #
    # Single item result, no response id
    myXml = "<db_response><status>OK</status><result>This is a result</result></db_response>"
    myResponse = dbResponse(myXml)
    myResponse.report()
    # Standard example, single result
    myXml = "<db_response><status>OK</status><result>True</result><response_id>request#1</response_id></db_response>"
    myResponse = dbResponse(myXml)
    myResponse.report()
    # List with multiple items
    myXml = "<db_response><status>OK</status><result><list><item>item1</item><item>item2</item></list></result><response_id>request#1</response_id></db_response>"
    myResponse = dbResponse(myXml)
    myResponse.report()
    # List with empty items
    myXml = "<db_response><status>OK</status><result><list><item></item><item></item></list></result><response_id>request#1</response_id></db_response>"
    myResponse = dbResponse(myXml)
    myResponse.report()
    # Example of broadcast unwrapping
    myXml = "<db_broadcast><message>Dummy broadcast message</message></db_broadcast>"
    myResponse = dbResponse(myXml)
    myResponse.report()
    # Example of unrecognised XML
    myXml = "<db_command>This should be unrecognised</db_command>"
    myResponse = dbResponse(myXml)
    myResponse.report()
