#
#     Copyright (C) 2006-7
#
#     This code is distributed under the terms and conditions of the
#     CCP4 Program Suite Licence Agreement as a CCP4 Library.
#     A copy of the CCP4 licence can be obtained by writing to the
#     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
#
#====================================================================
# CCP4 Interface - dbconsole.py
#
# An interactive "console" client application for dbccp4i
#
# Allow the user to issue commands to the database via
# the dbClientAPI.
#
# Usage:
# > tclsh[8.*] dbconsole.tcl
# 
# Peter Briggs
#
#====================================================================#
#
#Cvs_Id $Id: dbconsole.tcl,v 1.9 2007/01/25 11:58:26 pjx Exp $
#--------------------------------------------------------------#

# Acquire the external database procedures from the CCP4i core
#source [file join $env(CCP4I_TOP) src database.tcl]
#source [file join $env(CCP4I_TOP) src dbClientAPI.tcl]
source [file join $env(BIOXHIT_TOP) dbccp4i ClientAPI dbClientAPI.tcl]

# Application globals
set ignore_broadcasts 0

#################################################################
# Application procedures
#################################################################

#----------------------------------------------------------------
proc get_available_commands { filen } {
#----------------------------------------------------------------
    # Make a list of commands that are acceptable
    if { ![file exists $filen] } {
	return {}
    }
    set cmds {}
    set fp [open $filen "r"]
    while { ![eof $fp] } {
	gets $fp line
	set line [string trim $line]
	if { [regexp -- "^proc (\[A-Za-z0-9\]*) " $line full name] } {
	    lappend cmds $name
	    puts "Command: $name"
	}
    }
    close $fp
    return $cmds
}

#----------------------------------------------------------------
proc handle_command_line_input { allowed } {
#----------------------------------------------------------------
    # Read commands typed in at the command line
    # and farm them out to the processing function
    if { ![eof stdin] } {
	gets stdin cmd
    } else {
	set cmd "exit"
    }
    process_command $cmd $allowed
    return
}

#----------------------------------------------------------------
proc process_command { cmd allowed } {
#----------------------------------------------------------------
    global env
    # Process an API command line
    if { [lsearch [list "stop" "quit" "exit"] $cmd] > -1 } {
	# User wants to finish
	fileevent stdin readable ""
	set ::termination 1
    } elseif { [regexp "^ *list" $cmd] } {
	# List commands again
	get_available_commands \
	    [file join $env(BIOXHIT_TOP) dbccp4i ClientAPI dbClientAPI.tcl]
    } elseif { [regexp "^ *wait" $cmd] } {
	# Wait for a period of time (ms)
	if { [llength $cmd] == 2 } {
	    set wait [lindex $cmd 1]
	    puts "dbconsole: waiting for $wait ms..."
	    after $wait
	    puts -nonewline "...done\n% "
	} else {
	    puts -nonewline "dbconsole: wait instruction should be \"wait <time>\"\n% "
	}
    } elseif { [regexp "^ *hide_xml" $cmd] } {
	# Hide the XML data passing between the handler and the client
	DbXMLSetDebug 0
	puts -nonewline "dbconsole: echoing of XML messages turned OFF\n% "
    } elseif { [regexp "^ *show_xml" $cmd] } {
	# Show the XML data passing between the handler and the client
	DbXMLSetDebug 1
	puts -nonewline "dbconsole: echoing of XML messages turned ON\n% "
    } elseif { [regexp "^ *ignore_broadcasts" $cmd] } {
	# Ignore broadcast messages from the handler
	global ignore_broadcasts
	set ignore_broadcasts 1
	puts -nonewline "dbconsole: reporting of broadcast messages turned OFF\n% "
    } elseif { [regexp "^ *report_broadcasts" $cmd] } {
	# Report broadcast messages from the handler
	global ignore_broadcasts
	set ignore_broadcasts 0
	puts -nonewline "dbconsole: reporting of broadcast messages turned ON\n% "
    } elseif { [regexp "^ *DbShutdown" $cmd] } {
	# Special action for DbShutdown
	# Stop reading from stdin
	fileevent stdin readable ""
	# Send the shutdown instruction
	puts "dbconsole: sending shutdown command to handler..."
	catch { eval $cmd } msg
	puts "$cmd\nResult:\n$msg"
	# Terminate the console
	puts "dbconsole: exiting"
	set ::termination 1
    } elseif { [regexp "^ *@(.*)" $cmd m filen] } {
	# Line starts with @ symbol - indicates that commands
	# should be read from a file instead of stdin
	puts "dbconsole: read commands from script file \"$filen\""
	readfromfile $filen $allowed
    } elseif { [regexp "^ *\#" $cmd] } {
	# Comment line
	puts -nonewline "dbconsole: comment line ignored\n% "
    } else {
	# Check if this is an allowed command
	if { [regexp -- "^\[ \t\n\]*\$" $cmd] } {
	    puts -nonewline "\n% "
	    return
	}
	set cmd1 [lindex $cmd 0]
	if { [lsearch $allowed $cmd1] > -1 } {
	    catch { eval $cmd } msg
	    puts -nonewline "$cmd\nResult:\n$msg\n% "
	} else {
	    puts -nonewline "$cmd1: not an allowed command\n% "
	}
    }
}

#----------------------------------------------------------------
proc readfromfile { filen allowed } {
#----------------------------------------------------------------
    # Given a filename containing a "script" of API commands
    # execute these one at a time
    if { ![file exists $filen] } {
	# Duff file
	puts "dbconsole: error reading from file $filen (file not found)"
	return
    }
    set fp [open $filen "r"]
    while { [gets $fp cmd] > -1 } {
	puts "From script: $cmd"
	process_command $cmd $allowed
    }
    close $fp
    puts -nonewline "dbconsole: end of script file\n% "
    return
}   

#----------------------------------------------------------------
proc handlebroadcast { broadcast } {
#----------------------------------------------------------------
    # Handle broadcast message from the database handler
    global ignore_broadcasts
    if { ! $ignore_broadcasts } {
	puts "******************************************************"
	puts "Received broadcast message:\n\n$broadcast"
	puts "******************************************************"
	puts -nonewline "\n% "
    }
}

#################################################################
# Top-level script below here
#################################################################

# Preamble
puts "=========================================================="
puts " dbconsole for pythonic handler"
puts ""
puts " This is a \"console\" client application for the database"
puts " handler. It allows you to enter commands from the"
puts " clientAPI and see the responses from the server."
puts ""
puts " Built-in console commands:"
puts " wait <ms> : pause the console for <ms> milliseconds"
puts " list      : list the available commands"
puts " report_broadcasts : report broadcast messages (default)"
puts " ignore_broadcasts : don't report broadcast messages"
puts " hide_xml : suppress echoing of XML messages (default)"
puts " show_xml : echo XML messages between handler and client"
puts ""
puts " Type stop, quit or exit to exit the console application."
puts "=========================================================="
puts ""

# Start/connect to the database server
if { ![DbStartHandler handlebroadcast] } {
    puts "Failed to start the handler"
    exit 1
}
if { ![DbHandlerConnect] } {
    puts "Failed to connect to the handler"
    exit 1
}

# Get the available commands
set command_set [get_available_commands \
		     [file join $env(BIOXHIT_TOP) dbccp4i ClientAPI dbClientAPI.tcl]]
# Start the event handler
# Read commands from stdin and send them to the handler
puts -nonewline "Enter commands:\n% "
fconfigure stdin -blocking 0 -buffering line
fileevent stdin readable "handle_command_line_input \"$command_set\""
vwait termination

# Disconnect
puts "Disconnecting before exit, please wait..."
DbHandlerDisconnect

# Finished
puts "=========================================================="
puts " Finished"
puts "=========================================================="
exit
