#     Copyright (C) 2006 
#
#     This code is distributed under the terms and conditions of the
#     CCP4 licence agreement as `Part 1' (Annex 2) software.
#     A copy of the CCP4 licence can be obtained by writing to the
#     CCP4 Secretary, Daresbury Laboratory, Warrington WA4 4AD, UK.
#
#====================================================================
#
# CCP4 Interface - ccp4i.py
#
# Python module providing a class library for emulating the CCP4i
# def file based database. It also provides classes for dealing with
# the CCP4i "directories.def" file and general classes for .def files
# and lockfiles.
#
# Peter Briggs
#
########################################################################

__version__ = "$Revision: 1.23 $"

#######################################################################
# Import modules that this module depends on
#######################################################################
import sys
import os
import stat
import re
import time
import copy

#######################################################################
# Class definitions
#######################################################################

# database
#
# A class for handling CCP4i def file based databases
#
class database:
    """CCP4i def file project database class.

    Creates an instance of a CCP4i project database object associated
    with a particular project name and a project directory. If the
    project doesn't exist then the create method must be used to make
    a new project database.def file; use the open method to load
    data from an existing database."""

    def __init__(self,project,directory,template=""):
        """Internal: initialise a new project database object.

        The name of the project and the path to the corresponding
        project directory must be supplied.
        By default the data items in the file are taken from the
        'template' def file in $CCP4/ccp4i/etc; this can be over-ridden
        by specifying a template def file via the optional 'template'
        argument.
        No data is loaded until either the open or create methods
        are invoked."""
        self.__project = project
        self.__directory = GetAbsolutePath(directory)
        self.__record_template = []
        self.__joblist = []
        if template == "":
            # Default template
            self.__initfile = SearchPath("TOP","etc","database.def")
        else:
            # User defined template
            self.__initfile = template
        self.__database = array(initfile=self.__initfile)
        self.__dbfile = GetDbFile(self.__directory)
        # Make sure that the core db items are present
        for item in self.__list_core_db_items():
            if self.__database.addparameter(item,index=0,type=""):
                print "Added item "+item+" to the db"
        self.__dblock = dblockfile(project,self.__directory)
        self.__loaded = False
        self.__modified = False
        self.__auto_refresh = False
        self.__last_saved = time.time()
        self.__lastmessage = ""

    def __repr__(self):
        """Returns the project name."""
        return self.__project

    def __del__(self):
        """del(project)

        Saves data, releases lock and deletes the object."""
        self.save()
        if self.__dblock.haslock():
            self.__dblock.unlock()

    def __len__(self):
        """len(project)

        Returns the number of jobs stored in the project, or
        zero if the project is not loaded."""
        if self.isloaded():
            return len(self.__joblist)
        return 0

    def __nonzero__(self):
        """Returns True if the project is loaded, False otherwise."""
        return self.isloaded()

    def __initjoblist(self):
        """Internal method: initialises an internal list of jobs.

        Called when the project is opened, as part of the
        initialisation. This method goes through all possible job ids
        (from 1 to NJOBS) and checks for the existence of the 'core'
        data items. If some items are missing then it will attempt to
        repair the record by restoring the missing items with null
        values. If all items are missing then it is assumed that
        the job doesn't exist."""
        # Initialise the internal list of jobs
        self.__joblist = []
        n = self.njobs()
        maxjob = 0
        items = self.__list_core_db_items()
        # Range excludes the last item so add one to last job
        for job in range(1,n+1):
            jobexists = True
            nitems = 0
            missing_items = []
            for item in items:
                if not self.itemexists(job,item):
                    missing_items.append(item)
                else:
                    nitems =+ 1
            if nitems == 0:
                # None of the items were found
                jobexists = False
            else:
                # Some items were found, attempt to repair
                for item in missing_items:
                    self.__repairdata(job,item,"")
                    self.__modified = True
            if jobexists:
                # Add to the list of jobs
                self.__joblist.append(job)
                if job > maxjob:
                    maxjob = job
        if self.njobs() > maxjob:
            # Special case: the last job doesn't exist
            # Then we need to reset njobs
            self.__database["NJOBS"] = maxjob
            self.__modified = True
        return self.__joblist

    def __repairdata(self,job,item,value):
        """Internal method: force a data item to be set for a job.

        This should only be used internally on startup."""
        self.__database.setvalue(item,job,value)

    def __make_db_file(self,dbfile,project,dir,defid):
        """Internal: creates a new (empty) database.def file.

        dbfile is the absolute pathname specifying the database.def
        file; project is the name of the project, dir is the path
        of the project directory, and defid is the identifier for the
        type of data being written (probably should be 'database')."""
        # Make a new CCP4i database.def file.
        # The arguments are the absolute filename and the name of the
        # project. A database file will be generated from the template
        # database.def file in the CCP4 distribution
        self.__database.write(dbfile,project=project+" "+dir,defname=defid)

    # Make a database record
    def __make_db_record(self,job_id):
        """Internal: create a blank database record for a job."""
        # Make a blank database record for the specified job id
        for item in self.__list_db_items():
            self.__database.setvalue(item,job_id,None)

    def __list_db_items(self):
        """Internal: return a 'template list' of data items.

        The template list is the names of all data items stored
        for each job in the project."""
        # Return a template list of the items in database record
        if len(self.__record_template) == 0:
            # Load the template from the parameter list
            # of the def file
            for parameter in self.__database.parameters():
                if self.__database.parameter_isindexed(parameter):
                    root = self.__database.parameter_root(parameter)
                    self.__record_template.append(root)
        # Return the existing template
        return copy.copy(self.__record_template)

    def __list_core_db_items(self):
        """Internal: return a 'minimal template list' of data items.

        The minimal template list is the names of those data items
        which absolutely must be present for each job in the project."""
        # Return a "minimal" template list of the items that should
        # be in a database record
        return ["TITLE", \
            "TASKNAME", \
            "DATE", \
            "STATUS", \
            "INPUT_FILES", \
            "INPUT_FILES_DIR", \
            "INPUT_FILES_STATUS", \
            "OUTPUT_FILES", \
            "OUTPUT_FILES_DIR", \
            "OUTPUT_FILES_STATUS", \
            "LOGFILE"]
        
    def getProjectname(self):
        """Return the name of the project."""
        return self.__project
    
    def exists(self):
        """Test if the project exists.

        Returns True if the project directory and database.def file both
        exist, and False otherwise."""
        # Test if the project exists
        project = self.__project
        directory = self.__directory
        dbdir = GetDbDir(directory)
        dbfile = self.__dbfile
        if DirExists(dbdir) and FileExists(dbfile):
            return True
        else:
            return False

    def islocked(self):
        """Test whether the project is locked.

        Returns True if the project database.def file has a lockfile
        from any process, and False otherwise."""
        # Test if the project is currently locked
        return self.__dblock.islocked()

    def haslock(self):
        """Test whether the project is locked by this process.

        Returns True if the project database.def file has a lockfile
        owned by this process, and False otherwise."""
        return self.__dblock.haslock()

    def isloaded(self):
        """Test whether the project data has been loaded.

        Returns True if the database object has been populated from
        the file on disk."""
        return self.__loaded

    def isstale(self):
        """Check if the data in the object is older than on disk.

        Return True if the modification time of the resource
        file is more recent than the last save.
        If so then this suggests that the data in this object may
        not accurately mirror the data in the persistent
        storage."""
        if os.stat(self.__dbfile)[stat.ST_MTIME] > self.__last_saved:
            # Modification time is more recent than
            # the last save - data in object could
            # out-of-date
            return True
        else:
            return False

    def isreadable(self):
        """Test if it is possible to get data from the object.

        Returns True if the database object has been loaded,
        and has data that is at least as recent as the persistent
        storage on disk."""
        if self.isstale():
            if self.__auto_refresh: return self.refresh()
            return False
        return self.isloaded()

    def iswriteable(self):
        """Test if it is possible to modify and save the data.

        Returns True if the database object has been loaded, owns
        the lock on the persistent storage on disk, and has data
        that is at least as recent as the persistent storage."""
        if self.isstale():
            return False
        if self.isloaded() and self.haslock():
            return True
        else:
            return False

    def create(self):
        """Make a new project.

        Creating a new project includes creating the project directory
        (if it doesn't already exist), and creating the database
        subdirectory and database.def file. The project is then opened
        and the project object is returned.
        The create operation will fail (returning False) if the project
        already exists."""
        # Make a new database object
        if self.isloaded():
            self.__setmessage("Error project is already loaded for")
            return False
        project = self.__project
        directory = self.__directory
        if self.exists():
            self.__setmessage("Error: directory "
                              +directory+" is already a project dir")
            return False
        if not DirExists(directory):
            # Make the project directory
            MakeDir(directory)
        # Make the database subdirectory and database.def file
        dbdir = GetDbDir(directory)
        MakeDir(dbdir)
        dbfile = GetDbFile(directory)
        self.__make_db_file(dbfile,project,directory,"CCP4_Project_Database")
        return self.open()

    def set_auto_refresh(self,auto_refresh):
        """Set the auto_refresh flag.

        'auto_refresh' takes a boolean value. Setting it to 'True'
        means that a read attempt on a 'stale' database (i.e. when
        the source database file is newer than the data in the
        object) will automatically invoke the 'refresh' method to
        load the data in read-only mode. Setting it to 'False'
        means that the 'refresh' method must be explicitly invoked."""

        self.__auto_refresh = auto_refresh
        return self.__auto_refresh

    def auto_refresh(self):
        """Return the value of the auto_refresh flag.

        The 'auto_refresh' feature means that read attempts on a
        'stale' database automatically cause the data to be
        refreshed from file in read-only mode. See the
        set_auto_refresh method for more information."""

        return self.__auto_refresh

    def open(self,grablock=False,readonly=False):
        """Open a project.

        Read the data from the project database and return the project
        object.

        The open operation will fail (return False) if the project is
        locked, if the project is already loaded, or if the project
        doesn't exist. Set the 'grablock' argument to True to override
        any existing lock and force loading of the database.

        Specifying 'readonly' as True populates the database object
        from the file on disk without attempting to lock the resource.
        However it will clear an existing lock, so the process may
        still own the lock even if it is in 'read-only' mode."""
        # Open an existing database object
        project = self.__project
        directory = self.__directory
        if self.isloaded():
            self.__setmessage("Error project is already loaded for"+project)
            return False
        dbfile = GetDbFile(directory)
        if not self.exists():
            self.__setmessage("Error: directory "+directory+" is not a project")
            return False
        # Deal with locks, if necessary
        if not readonly:
            if self.islocked():
                if not grablock:
                    self.__setmessage("Error: database file "+dbfile+
                                      " is locked by another process")
                    return False
                else:
                    self.__dblock.unlock(force=True)               
            self.__dblock.lock()
        # Read in the data
        self.__database.read(dbfile)
        # Get the project name from the def file header
        # FIXME these checks should result in some action if failed
        rx = re.compile(r"^PROJECT ([^ ]+) ([^ ]+)")
        check = rx.match(self.__database.getheaderdata("^PROJECT"))
        if check:
            chk_project = check.group(1)
            if project != chk_project:
                print "Warning: project name in file is "+str(chk_project)
            chk_directory = check.group(2)
            if directory != chk_directory:
                print "Warning: directory in file is "+str(chk_directory)
        else:
            print "Warning: failed to get project & dir names from def file"
        self.__last_saved = os.stat(dbfile)[stat.ST_MTIME]
        self.__loaded = True
        self.__joblist = self.__initjoblist()
        return True

    def refresh(self,grablock=False,readonly=True):
        """Reload the data from the file into a database object.

        Use the 'refresh' method to re-read the data from the
        file on disk into an existing (loaded) database object.
        This is necessary for example if the file is updated
        by an external process.

        Essentially the database object is closed and then
        reopened. By default the data is reloaded in 'readonly'
        mode. Set the 'grablock' argument to True to override
        any existing lock and force loading of the database in
        write mode - see the 'open' method."""
        self.close()
        self.__database = array(initfile=self.__initfile)
        self.__modified = False
        self.__last_saved = time.time()
        return self.open(grablock=grablock,readonly=readonly)

    def close(self):
        """Close an open project.

        This saves any unsaved data to file, releases the lock on the
        database.def file, and clears the data from the project object,
        returning True on success.
        The project must have first been loaded via an open or create
        operation, otherwise False is returned."""
        # Close an open database
        if not self.isloaded():
            self.__setmessage("Warning: database for "
                              +self.__project+" isn't loaded")
            return False
        self.save()
        if self.haslock():
            self.__dblock.unlock()
        self.__database = array()
        self.__record_template = []
        self.__joblist = []
        self.__loaded = False
        return True

    def selectjobs(self,item,pattern):
        """Retrieve a list of jobs based on some selection criterion.

        Returns a list of jobs for which the value of the given data
        item matches the supplied regular expression pattern."""
        # Return a list of jobs where the values of the
        # given item match the supplied regular expression
        if not self.isreadable():
            return []
        n = self.njobs()
        rx = re.compile(pattern)
        selection = []
        for job in self.__joblist:
            if self.itemexists(job,item):
                if rx.match(str(self.getdata(job,item))):
                    selection.append(job)
        return selection

    def njobs(self):
        """Return the value of the NJOBS data item.

        For historical reasons, the NJOBS data item records the current
        highest job id number used in the project, and is not generally
        the actual number of jobs in the project. To get the actual
        number of jobs in an object project, do len(project).
        Returns -1 if the project has not been correctly loaded."""
        # Return the value of NJOBS
        if not self.isreadable():
            return -1
        return int(self.__database["NJOBS"])

    def listjobs(self):
        """Return the list of all jobs in the project.

        The list is an unsorted list of job id numbers, or an empty list
        if the project is not currently loaded."""
        # Return the list of job ids
        if not self.isreadable():
            return []
        # Don't return the list itself, as it is mutable
        return copy.copy(self.__joblist)

    def newjob(self,taskname="",status="",title=""):
        """Create a new job record in the project.

        Creates a new entry in the current project with the current date
        and returns the id number for the new job. Optionally the
        taskname (TASKNAME data item), the status (STATUS item) and/or
        title (TITLE item) may also be set when the job is created.
        The operation will fail and return -1 if project is not loaded,
        or if the current object has lost the lock on the database.def
        file."""
        # Create a new record in the database
        if not self.iswriteable():
            self.__setmessage("Error: (newjob) project is not writeable")
            return -1
        n = self.njobs() + 1
        self.__make_db_record(n)
        self.__database["NJOBS"] = n
        self.__joblist.append(n)
        # Optionally set the attributes of the job
        self.setdata(n,"TASKNAME",taskname)
        self.setdata(n,"STATUS",status)
        self.setdata(n,"TITLE",title)
        # Set the date
        self.updatetime(n)
        self.__modified = True
        return n

    def deletejob(self,job):
        """Remove an existing job record from the project.

        The job with the specified id number will be removed from the
        project, also removing all associated data, returning True on
        success. The operation will fail if the database is not loaded,
        if the object has lost the lock on the database.def file, or
        if a job with the specified id number is not found."""
        # Remove a record from the database
        if not self.iswriteable():
            self.__setmessage("Error: (deletejob) project is not writeable")
            return False
        n = self.njobs()
        if job > n:
            return False
        # Locate associated items in the database
        ndel = 0
        for item in self.__list_db_items():
            if self.__database.has_parameter(item,job):
                self.__database.delparameter(item,job)
                ndel = ndel + 1
        if ndel == 0:
            # Nothing was deleted
            return False
        if job == n:
            self.__database["NJOBS"] = n - 1
        self.__joblist.remove(job)
        self.__modified = True
        return True

    def itemexists(self,job,item):
        """Check for the existence of a data item for a particular job.

        Returns True if the data item is found for the specified job
        id number, and False if not. Also returns False in the event
        that the project is not loaded."""
        # Confirm whether or not an item exists for a job
        if not self.isreadable():
            self.__setmessage("itemexists: project is not readable")
            return False
        if job > self.njobs():
            self.__setmessage("itemexists: job > "+str(self.njobs()))
            return False
        return self.__database.has_parameter(item,job)

    def setdata(self,job,item,value):
        """Set the value of a data item for the specified job.

        The job is specified as an id number, and the operation returns
        True on success, and False on failure. The operation will fail
        if: the project is not loaded; the data item doesn't exist for
        the job; the object has lost the lock on the database.def file.
        No attempt is made to check the validity of the new value."""
        # Set the value of an item for a job
        if not self.iswriteable():
            self.__setmessage("setdata: project is not writeable")
            return False
        if not self.itemexists(job,item):
            raise IndexError, \
                  "Item \""+str(item)+"\" doesn't exist for job \""+str(job)+"\""
        self.__database.setvalue(item,job,value)
        self.__modified = True
        return True

    def updatetime(self,job):
        """Update the time of the job to be the current time.

        This wraps the setdata method to automatically set the DATE
        attribute of the job specified by the id number, and returns
        the result of the setdata operation."""
        # Update the DATE attribute of a job to be
        # the current date
        if self.iswriteable():
            return self.setdata(job,"DATE",int(time.time()))
        return False

    def getdata(self,job,item):
        """Retrieve the value of a data item stored for a job.

        The job is specified by an id number. If the item doesn't exist
        (established through a call to the itemexists method) then the
        operation raise an IndexError exception - otherwise the value of
        the data item is returned."""
        # Retrieve the value of an item
        if not self.isreadable():
            return None
        if not self.itemexists(job,item):
            raise IndexError, \
                  "Item \""+str(item)+"\" doesn't exist for job \""+str(job)+"\""
        return self.__database.getvalue(item,job)

    def describe(self,joblist,itemlist,formatlist):
        """Return a list of formatted strings populated with job data.

        For each job id number in the joblist, the describe method
        retrieves the values of each of the data items in the itemlist
        and creates a description string by placing those values into
        fields with character widths supplied in the formatlist.
        It returns a list of these strings, with one list item per job.
        Allowed data items can be any stored in the project database
        plus JOB_ID (the id number of the job), which is an implicit
        data item."""
        # Return a list of descriptions, one item per job
        # from the supplied list of jobs
        description = []
        if not self.isreadable():
            return description
        for job in joblist:
            description.append(self.describejob(int(job),itemlist,formatlist))
        return description

    def describejob(self,job,itemlist,formatlist):
        """Return a formatted string populated with job data.

        For the specified job id number, the describejob method
        retrieves the values of each of the data items in the itemlist
        and creates a description string by placing those values into
        fields with character widths supplied in the formatlist.
        Allowed data items can be any stored in the project database
        plus JOB_ID (the id number of the job), which is an implicit
        data item."""
        # Write out the description of a single job record
        # itemlist is a list of data items to be written
        # formatlist is a corresponding list of fieldwidths
        if not self.isreadable():
            return []
        line_format = ""
        for field in formatlist:
            line_format = line_format + "%-" + str(field) + "s "
        valuelist = []
        for item in itemlist:
            if item == "JOB_ID":
                # Job id is not explicity stored
                value = str(job)
            elif item == "DATE":
                # Date must be converted to a pretty format
                now = time.time()
                epoch = int(self.getdata(job,"DATE"))
                jobtime = time.localtime(epoch)
                if (now - epoch) < 86400:
                    # Within 24 hours of the current time
                    value = time.strftime("%H:%M:%S",jobtime)
                else:
                    # Older than 24 hours
                    value = time.strftime("%d %b %y",jobtime)
            else:
                value = str(self.getdata(job,item))
            valuelist.append(value)
        valuelist = tuple(valuelist)
        return line_format % valuelist

    def save(self):
        """Save the project database contents to file.

        The save method writes the data in the object to persistent
        storage (i.e. the database.def file for the project).
        On successful completion True is returned.
        The save operation will fail if the project is not loaded,
        or if the object has lost the lock on the database.def file;
        in both cases False is returned."""
        # Save the database contents to file
        if not self.__loaded:
            return False
        # Check whether the disk storage is newer
        if self.isstale():
            self.__setmessage("Error: (save) data on disk is newer than in memory")
            return False
        # Check the lock
        if not self.__dblock.haslock():
            self.__setmessage("Error: (save) lock has been lost")
            return False
        if not self.__modified:
            # No need to write a file
            return True
        dbfile = GetDbFile(self.__directory)
        self.__database.write(dbfile,
                              project=self.__project+" "+self.__directory,
                              defname="CCP4_Project_Database")
        self.__last_saved = os.stat(dbfile)[stat.ST_MTIME]
        self.__modified = False
        return True

    def __setmessage(self,message):
        """Internal: set the error message from the last failed operation."""
        # Error reporting
        # Other methods in the class can set an error message
        self.__lastmessage = str(message)
        return True

    def getmessage(self):
        """Retrieve last internal error message.

        If an operation fails then the client application can
        access the last error message using this method. Invoking
        getmessage clears the error message."""
        # Error reporting
        # Application can access the last error message
        message = self.__lastmessage
        self.__lastmessage = ""
        return message

# lockfile
#
# A class for handling file locks
#
class lockfile:
    """CCP4i lock file class.

    This class provides methods for creating and manipulating lock
    files associated with arbitrary files within CCP4i."""

    def __init__(self,filen,resource_desc=""):
        self.__filen = filen
        self.__rootname = GetFileRootname(filen)
        self.__lockfile = os.path.join(self.__rootname+".LOCK")
        if resource_desc == "":
            self.__resource_desc = os.path.basename(self.__rootname)
        else:
            self.__resource_desc = resource_desc
        self.__user = GetUserId()
        self.__date = GetDate()
        self.__pid = str(GetPid())

    def resource(self):
        # Return the name of the resource (i.e. file) being locked
        return self.__filen

    def filename(self):
        # Return the lockfile name (including the path)
        return self.__lockfile

    def islocked(self):
        # Check whether the database is locked
        return os.path.isfile(self.filename())

    def haslock(self):
        # Check whether this lock object owns the lockfile
        if not self.islocked():
            return False
        # Read the lockfile
        try:
            f = open(self.filename(),"r")
        except:
            # Catch all errors
            ##print "Unable to open lockfile for reading"
            # Pass the error up to the calling application
            raise
        # Read in the contents of the lockfile
        contents = f.read()
        # Create a regular expression for picking out information
        rx = re.compile(r"^Lock file for (.*)\nCreated by (.*)\nDate (.*)\nOwned by pid (.*)\n$")
        check = rx.match(contents)
        if not check:
            # Failed to match
            ##print "Error lock file format doesn't match regular expression"
            return False
        else:
            # Extract the attributes
            resource_desc = check.group(1)
            if self.__resource_desc != resource_desc:
                ##print "Lock file: resource descr doesn't match"
                return False
            user = check.group(2)
            if self.__user != user:
                ##print "Lock file: user name doesn't match"
                return False
            date = check.group(3)
            if self.__date != date:
                ##print "Lock file: time/date doesn't match"
                return False
            pid = check.group(4)
            if self.__pid != pid:
                ##print "Lock file: pid doesn't match"
                return False
            # All the tests are passed
            return True

    def lock(self):
        # Lock the file resource
        if self.islocked():
            print "Error: resource '"+str(self.resource())+"' is already locked"
            return False
        # Create a lockfile
        try:
            f = open(self.filename(),"w")
        except:
            # Catch all errors
            print "Unable to open lockfile for writing"
            # Pass the error up to the calling application
            raise
        # Write the contents of the lockfile
        # The lockfile conists of the following lines:
        # Lock file for <project/resource description>
        # Created by <uid>
        # Date <date>
        # Owned by pid <pid>
        f.write("Lock file for "+self.__resource_desc+"\n")
        f.write("Created by "+str(self.__user)+"\n")
        f.write("Date "+str(self.__date)+"\n")
        f.write("Owned by pid "+str(self.__pid)+"\n")
        f.close()
        return 1
        
    def unlock(self,force=False):
        # Unlock the file resource by removing the lockfile
        # Setting force to be true removes any lock
        if not self.islocked():
            print "Error: resource '"+str(self.resource())+"' is not currently locked"
            return 0
        if not self.haslock() and not force:
            print "Error: lockfile doesn't belong to this process"
            return 0
        # Delete the lockfile
        os.remove(self.filename())
        return True

# dblockfile
#
# A class specifically for handling database lockfiles
# This inherits from the general CCP4i lockfile class
#
class dblockfile(lockfile):
    """CCP4i database lock file class.

    This class provides methods for creating and manipulating lock
    files for CCP4i def file based databases."""

    def __init__(self,project,directory):
        filen = os.path.join(GetDbDir(directory),"database.def")
        lockfile.__init__(self,filen,resource_desc=project)

# directories
#
# A class for handling CCP4i directories.def data
#
class directories:
    """CCP4i project directories class.

    This class provides methods for reading and writing the project
    directory information in directories.def for CCP4i-based
    applications."""

    def __init__(self,user="",source=""):
        if user == "":
            self.__user = GetUserId()
        else:
            self.__user = user
        if source == "":
            self.__directories = os.path.join(GetDotCCP4(),GetOpsys(),"directories.def")
        else:
            self.__directories = source
        self.__projects = {}
        self.__db_dirs = {}
        self.__def_dirs = {}
        # Initialise from the default distribution file
        self.__data = array(SearchPath("TOP","etc","directories.def.dist"))
        self.__lock = lockfile(self.__directories)
        self.__loaded = False
        self.__modified = False
        self.__auto_refresh = False
        self.__last_saved = time.time()
        self.__lastmessage = ""

    def __repr__(self):
        if self.__loaded:
            return self.__user
        else:
            return ""

    def __nonzero__(self):
        return self.__loaded

    def __del__(self):
        if self.__loaded:
            self.release()

    def user(self):
        # Return the username
        return self.__user

    def source(self):
        # Return the source filename
        return self.__directories

    def iswriteable(self):
        # Test if it is possible to modify and save
        if self.islockedbyccp4i() or self.isstale():
            return False
        if self.isloaded() and self.haslock():
            return True
        else:
            return False

    def isreadable(self):
        # Test if it is possible to get data from the class
        if self.isstale():
            # The image in the object is older than
            # the file on disk
            # Attempt to refresh the data
            if self.__auto_refresh: return self.refresh()
            return False
        return self.isloaded()

    def isloaded(self):
        # Test if the data has been loaded from the file
        return self.__loaded

    def islocked(self):
        # Test if the directories file is currently locked
        return self.__lock.islocked()

    def haslock(self):
        # Return True if the process owns the lock
        return self.__lock.haslock()

    def isstale(self):
        # Return True if the modification time of the resource
        # file is more recent than the last save.
        # This suggests that the data in this object may
        # not accurately mirror the data in the persistent
        # storage.
        if os.stat(self.__directories)[stat.ST_MTIME] > self.__last_saved:
            # Modification time is more recent than
            # the last save - data in object could
            # out-of-date
            return True
        else:
            return False

    def islockedbyccp4i(self):
        # Return True if the directories resource is also
        # locked by CCP4i
        # CCP4i's lock file is called <filen>.LOCK, and
        # CCP4i does not respect lock files created on
        # behalf of this class.
        #
        # Caution: CCP4i's lock on directories.def for
        # 1.4.4 seems a little unpredictable. So it
        # may be that this returns "False" even if
        # CCP4i is running.
        ccp4ilock = str(self.__directories)+".LOCK"
        if FileExists(ccp4ilock):
            return True
        return False
    
    def set_auto_refresh(self,auto_refresh):
        """Set the auto_refresh flag.

        'auto_refresh' takes a boolean value. Setting it to 'True'
        means that a read attempt on a 'stale' database (i.e. when
        the source database file is newer than the data in the
        object) will automatically invoke the 'refresh' method to
        load the data in read-only mode. Setting it to 'False'
        means that the 'refresh' method must be explicitly invoked."""

        self.__auto_refresh = auto_refresh
        return self.__auto_refresh

    def auto_refresh(self):
        """Return the value of the auto_refresh flag.

        The 'auto_refresh' feature means that read attempts on a
        'stale' database automatically cause the data to be
        refreshed from file in read-only mode. See the
        set_auto_refresh method for more information."""

        return self.__auto_refresh

    def create(self):
        # Make a new directories file
        directories = self.__directories
        if not FileExists(directories):
            # The file must not already exist
            array = self.__data
            array.write(directories)
            # Perform load operations
            return self.load()
        # Unable to create new file
        return False

    def load(self,force=False,readonly=False):
        # Load the data for the user
        # Setting force to True over-rides an existing lock
        # Setting readonly to True loads the data even if
        # there is a lock from another process
        if self.isloaded():
            return False
        directories = self.__directories
        if FileExists(directories):
            # Deal with locks
            if not readonly and self.islocked():
                if force:
                    self.__lock.unlock(force=True)
                else:
                    return False
            if not readonly and not self.islocked():
                self.__lock.lock()
            # Read in the data
            array = self.__data
            array.read(self.__directories)
            # Make local copy of project alias/directory lookup
            nprojects = int(array["N_PROJECTS"])
            for alias in array.listindexed("PROJECT_ALIAS"):
                path = array.getequivalent(alias,"PROJECT_PATH")
                db_dir = array.getequivalent(alias,"PROJECT_DB")
                self.__projects[array[alias]] = array[path]
                self.__db_dirs[array[alias]] = array[db_dir]
            # Make local copy of def_dir alias/directory lookup
            ndefdirs = int(array["N_DEF_DIRS"])
            for alias in array.listindexed("DEF_DIR_ALIAS"):
                path = array.getequivalent(alias,"DEF_DIR_PATH")
                self.__def_dirs[array[alias]] = array[path]
            self.__loaded = True
            # Set last saved time to the last modification time
            # of the file now
            self.__last_saved = os.stat(directories)[stat.ST_MTIME]
            return True
        else:
            # Directories file didn't exist
            return False

    def refresh(self,readonly=True):
        # Dump the contents of the directories object and
        # reread data from the file
        self.release()
        self.__projects = {}
        self.__db_dirs = {}
        self.__def_dirs = {}
        self.__data = array(SearchPath("TOP","etc","directories.def.dist"))
        return self.load(readonly=readonly)

    def release(self):
        # Release the resource and surrender the lock
        # Conditions on save: 1. must be loaded 2. must own the lock
        if self.iswriteable():
            self.save()
        if self.haslock():
            self.__lock.unlock()
        self.__loaded = False
        return True

    ### FIXME
    ### This function provided for diagnostic purposes
    def array(self):
        # Return the underlying array object
        if self.isreadable():
            return self.__data
        else:
            raise AttributeError, "Data not loaded"

    def logdir(self):
        # Return the value of LOG_DIRECTORY
        if self.isreadable():
            return self.__data["LOG_DIRECTORY"]
        else:
            raise AttributeError, "Data not loaded"

    def setlogdir(self,log_dir):
        # Set LOG_DIR to a new value
        if self.iswriteable():
            self.__data["LOG_DIRECTORY"] = log_dir

    def project_menu(self):
        # Return the value of PROJECT_MENU
        if self.isreadable():
            return self.__data["PROJECT_MENU"]
        else:
            raise AttributeError, "Data not readable"

    def listprojects(self):
        # Return a list of the user's projects
        if self.isreadable():
            return self.__projects.keys()
        else:
            raise AttributeError, "Data not readable"
        
    def n_projects(self):
        #Return the number of projects
        if self.isreadable():
            return int(self.__data["N_PROJECTS"])
        else:
            raise AttributeError, "Data not readable"

    def projectdir(self,project):
        # Return the directory corresponding to a project name
        if self.isreadable():
            if self.__projects.has_key(project):
                return self.__projects[project]
            else:
                return ""
        else:
            raise AttributeError, "Data not readable"

    def projectdb(self,project):
        # Return the CCP4_DATABASE directory corresponding
        # to a project name
        if self.isreadable():
            if self.__db_dirs.has_key(project):
                return self.__db_dirs[project]
            else:
                return ""
        else:
            raise AttributeError, "Data not readable"

    def isproject(self,project):
        # Determine whether or not the project name corresponds
        # to an existing project
        return self.__projects.has_key(project)

    def addprojectref(self,alias,path,db_dir=""):
        # Add a new project reference
        if self.iswriteable():
            if not self.__projects.has_key(alias):
                array = self.__data
                abspath = GetAbsolutePath(path)
                self.__projects[alias] = abspath
                if db_dir != "":
                    self.__db_dirs[alias] = db_dir
                else:
                    # Determine from directory name
                    self.__db_dirs[alias] = GetDbDir(abspath)
                n = array.incr("N_PROJECTS")
                array.setvalue("PROJECT_ALIAS",n,alias)
                array.setvalue("PROJECT_PATH",n,abspath)
                array.setvalue("PROJECT_DB",n,db_dir)
                # Update the PROJECT_MENU parameter
                self.__update_project_menu()
                self.__modified = True
                return True
            else:
                # Project reference already exists
                return False
        else:
            return False

    def deleteprojectref(self,alias):
        # Remove a project reference
        if self.iswriteable():
            if self.__projects.has_key(alias):
                array = self.__data
                n = array["N_PROJECTS"]
                for project in array.listindexed("PROJECT_ALIAS"):
                    if array[project] == alias:
                        # Located the project reference
                        i = array.parameter_index(project)
                        if i == n:
                            array.delparameter("PROJECT_ALIAS",n)
                            array.delparameter("PROJECT_PATH",n)
                            array.delparameter("PROJECT_DB",n)
                        else:
                            array.replacevalue("PROJECT_ALIAS",i,n)
                            array.replacevalue("PROJECT_PATH",i,n)
                            array.replacevalue("PROJECT_DB",i,n)
                        array.incr("N_PROJECTS",-1)
                        self.__modified = True
                        # Update local data
                        del self.__projects[alias]
                        del self.__db_dirs[alias]
                        # Update the PROJECT_MENU parameter
                        self.__update_project_menu()
                        return True
            # Didn't find the project
            return False
        else:
            return False

    def listdefdirs(self):
        # Return a list of the user's default directories
        if self.isreadable():
            return self.__def_dirs.keys()
        
    def n_defdirs(self):
        #Return the number of def dirs
        if self.isreadable():
            return int(self.__data["N_DEF_DIRS"])
        else:
            raise AttributeError, "Data not readable"

    def defdir(self,def_dir):
        # Return the directory corresponding to a def dir name
        if self.isreadable():
            if self.__def_dirs.has_key(def_dir):
                return self.__def_dirs[def_dir]

    def isdefdir(self,def_dir):
        # Determine whether or not the def dir name corresponds
        # to an existing def dir
        if self.isreadable():
            return self.__def_dirs.has_key(def_dir)

    def adddefdirref(self,def_dir,path):
        # Add a new def dir reference
        if self.iswriteable():
            if not self.__def_dirs.has_key(def_dir):
                array = self.__data
                abspath = GetAbsolutePath(path)
                self.__def_dirs[def_dir] = abspath
                n = array.incr("N_DEF_DIRS")
                array.setvalue("DEF_DIR_ALIAS",n,def_dir)
                array.setvalue("DEF_DIR_PATH",n,abspath)
                self.__modified = True
                return True
            else:
                # Def dir reference already exists
                return False
        else:
            return False

    def deletedefdirref(self,def_dir):
        # Remove a def dir reference
        if self.iswriteable():
            if self.__def_dirs.has_key(def_dir):
                array = self.__data
                n = array["N_DEF_DIRS"]
                for name in array.listindexed("DEF_DIR_ALIAS"):
                    if array[name] == def_dir:
                        # Located the project reference
                        i = array.parameter_index(name)
                        if i == n:
                            array.delparameter("DEF_DIR_ALIAS",n)
                            array.delparameter("DEF_DIR_PATH",n)
                        else:
                            array.replacevalue("DEF_DIR_ALIAS",i,n)
                            array.replacevalue("DEF_DIR_PATH",i,n)
                        array.incr("N_DEF_DIRS",-1)
                        self.__modified = True
                        # Update local data
                        del self.__def_dirs[def_dir]
                        return True
            # Didn't find the project
            return False
        else:
            return False

    def save(self,filen=""):
        # Save the data to disk
        if not self.__loaded:
            print "Not loaded"
            return False
        # Save to file
        if filen != "" and filen != self.__directories:
            self.__data.write(filen,include_types=True)
            return True
        elif self.__modified and self.__lock.haslock():
            self.__data.write(self.__directories,include_types=True)
            self.__modified = False
            self.__last_saved = os.stat(self.__directories)[stat.ST_MTIME]
            return True
        else:
            self.__setmessage("Error: unable to save directories data")
            return False

    def fullfilename(self,filen,project):
        # Given a filename filen and a project name
        # return a single full path
        if not self.__loaded:
            print "Not loaded"
            return ""
        if self.isproject(project):
            return os.path.join(self.projectdir(project),filen)
        elif self.isdefdir(project):
            return os.path.join(self.defdir(project),filen)
        else:
            return filen

    def __update_project_menu(self):
        # Private: update the PROJECT_MENU parameter when
        # projects are added or deleted
        project_menu = ""
        for project in self.listprojects():
            project_menu = project_menu+str(project)+" "
        self.__data["PROJECT_MENU"] = project_menu.strip()

    def __setmessage(self,message):
        # Error reporting
        # Other methods in the class can set an error message
        self.__lastmessage = str(message)
        return True

    def getmessage(self):
        # Error reporting
        # Application can access the last error message
        message = self.__lastmessage
        self.__lastmessage = ""
        return message

# history
#
# A class for generating and retrieving data flow information
# from a database
class history:
    """CCP4i database history object.

    Supplied with a loaded database object and a loaded directories
    object, use the construct method to populate the history object.
    Use the parentsof and childrenof methods to return a list of
    parent/child job id numbers for each job. Use the update method
    to find and add new relationships for an existing history.
    Use the arelinked method to find out if two job ids are linked."""

    def __init__(self,database,directories):
        self.__database = database
        self.__directories = directories
        self.__initialise_lists()

    def __initialise_lists(self):
        # Clear the existing lists and reset
        self.__parents = dict()
        self.__children = dict()
        if self.__database:
            for job in self.__database.listjobs():
                self.__parents[job] = []
                self.__children[job] = []

    def updatelinks(self,job_id):
        # (Re)generate the links between the specified
        # job and all other jobs
        # Not implemented yet
        return

    def construct(self):
        # Construct relationships between jobs in the
        # database from scratch
        if self.__database and self.__directories:
            # Initialise lists
            self.__initialise_lists()
            self.__populate()

    def update(self):
        # Update the history by looking for new links/jobs
        # without clearing existing ones
        if self.__database and self.__directories:
            self.__populate()

    def __populate(self):
        # Loop over jobs and generate links if necessary
        # Loop over all jobs in the project
        joblist = self.__database.listjobs()
        for i in range(0,len(joblist)):
            job_i = joblist[i]
            infiles_i = self.infiles(job_i)
            outfiles_i = self.outfiles(job_i)
            # Loop over subsequent jobs
            for j in range(i+1,len(joblist)):
                job_j = joblist[j]
                if self.arelinked(job_i,job_j):
                    # These jobs are already linked
                    continue
                # Otherwise explicitly check for links
                # by matching input and output
                infiles_j = self.infiles(job_j)
                outfiles_j = self.outfiles(job_j)
                # Do any input files match output files
                # for this job?
                for filei in infiles_i:
                    for filej in outfiles_j:
                        if str(filei) == str(filej):
                            # Found a link
                            self.addlink(job_j,job_i)
                            break
                    for filei in outfiles_i:
                        for filej in infiles_j:
                            if str(filei) == str(filej):
                                # Found a link
                                self.addlink(job_i,job_j)
                                break
        return

    def addlink(self,parent_job,child_job):
        # Record a link between two jobs
        try:
            x = self.__parents[child_job].index(parent_job)
        except:
            self.__parents[child_job].append(parent_job)
        try:
            x = self.__children[parent_job].index(child_job)
        except:
            self.__children[parent_job].append(child_job)

    def infiles(self,job_id):
        # Return a list of input files for this job
        infiles = self.__database.getdata(job_id,"INPUT_FILES").split()
        indirs = self.__database.getdata(job_id,"INPUT_FILES_DIR").split()
        filelist = []
        for k in range(0,len(infiles)):
            filelist.append(self.__directories.fullfilename(infiles[k],
                                                            indirs[k]))
        return filelist
        
    def outfiles(self,job_id):
        # Return a list of output files for this job
        infiles = self.__database.getdata(job_id,"OUTPUT_FILES").split()
        indirs = self.__database.getdata(job_id,"OUTPUT_FILES_DIR").split()
        filelist = []
        for k in range(0,len(infiles)):
            filelist.append(self.__directories.fullfilename(infiles[k],
                                                            indirs[k]))
        return filelist

    def arelinked(self,job_id1,job_id2):
        # Return True if the two jobs are currently
        # linked in the history object
        try:
            x = self.__parents[job_id1].index(job_id2)
            return True
        except:
            try:
                x = self.__children[job_id1].index(job_id2)
                return True
            except:
                return False

    def parentsof(self,job_id):
        # Return a list of job ids corresponding
        # to the "parent" jobs of the specified id
        return copy.copy(self.__parents[job_id])

    def childrenof(self,job_id):
        # Return a list of job ids corresponding
        # to the "children" jobs of the specified id
        return copy.copy(self.__children[job_id])

    def allchildrenof(self,job_id,result_ids):
        # Return a list of job ids of all of the children of the specific id
        children = self.childrenof(job_id)
        if children == []:
            return result_ids
        else:
            for child in children:
                if child not in result_ids:
                    result_ids.append(child)
                self.allchildrenof(child,result_ids)
        return result_ids

    def allparentsof(self,job_id,result_ids):
        # Return a list of job ids of all of the parent of the specific id
        parents = self.parentsof(job_id)
        if parents == []:
            return result_ids
        else:
            for parent in parents:
                if parent not in result_ids:
                    result_ids.append(parent)
                self.allparentsof(parent,result_ids)
        return result_ids
        
# array
#
# A class for dealing with CCP4i arrays, which are loaded from and
# saved to .def files
#
class array:
    """CCP4i array class.

    Read and write CCP4i .def files and allow manipulation of data."""

    def __init__(self,initfile=""):
        # Initialisation
        # initfile is an optional def file to initialise from
        self.__initfile = initfile
        # Dictionaries and lists to hold the contents
        self.__header = []
        self.__contents = {}
        self.__types = {}
        self.__parameters = []
        # Header info for writing out
        # Default application name 
        self.__application = "CCP4IPy "+version()
        # Initialise
        if self.__initfile != "":
            self.read(initfile)

    def __getitem__(self,key):
        # Built-in: allows array[key] to return value
        if self.__contents.has_key(key):
            return self.__contents[key]
        else:
            raise KeyError, "Key not found"

    def __setitem__(self,key,value):
        # Built-in: allows array[key] = value to set value
        self.__contents[key] = value

    def __delitem__(self,key):
        # Built-in: allows del array[key] to delete item
        del self.__contents[key]

    def __contains__(self,key):
        # Built-in: allows test key in array
        if self.__contents.has_key(key):
            return True
        # Check that this isn't an indexed parameter and the
        # calling subprogram didn't specify the index
        if str(key).find(",") < 0:
            return self.__contents.has_key(str(key)+",0")
        else:
            return False

    def read(self,filename):
        """Reads the contents of a def-format file into the array.

        The file format can be either key-value pairs, or
        key-type-value triplets. If a value in the file is of the
        form '\\$name' then 'name' is assumed to be an environment
        variable which is substituted via a call to the GetEnvVar
        function."""
        # Read the contents of a file into the
        # array object
        try:
            f = open(str(filename),"r")
        except:
            # Catch all errors and pass back up
            self.__error = "Failed to open file \"%s\"" % str(filename)
            raise

        # Initialise regular expressions for various things
        # ...CCP4i header lines
        header = re.compile(r'^\s*#CCP4I\s*(.*)$')
        # ...for comment lines that are ignored
        ignore = re.compile(r'^\s*#')

        # Read the file line-wise
        for line in f:
            # Remove trailing newline
            line = line.rstrip("\n")
            if header.match(line):
                # Store header info
                # FIXME: there is no resetting of header data read
                # from multiple files
                self.__header.append(header.match(line).group(1))
                continue
            if ignore.match(line):
                # Skip other lines starting with # symbol
                continue
            # Tokenise remaining lines
            # There are two types of line: parameter-value pairs
            # and parameter-type-value triples
            tokens = tokenise(line)
            if len(tokens) == 2 or len(tokens) == 3:
                # Break up into components
                key = tokens[0]
                if len(tokens) == 3:
                    type = tokens[1]
                else:
                    type = ""
                # Store as a parameter
                self.addparameter(key,type=type)
                # Store the value (less any quotes)
                value = str(tokens[-1]).strip("\"")
                if str(value).startswith("\\$"):
                    # The value has a leading \$
                    # which indicates that it's an
                    # environment variable
                    value = GetEnvVar(value[2:])
                self[key] = value
        # Finished reading - close the file
        f.close()

    def write(self,filename,
              application="",
              version="",
              defname="",
              user="",
              project="",
              include_types=False,
              include_prototypes=False):
        """Write the contents of the array object to a def file.

        If include_types is True then the type information will also be
        written to the file. If include_prototypes is True then the
        parameter prototypes for indexed parameters (i.e. the 0th-indexed
        parameters) will also be written."""
        # Write the array contents out to a .def file
        try:
            f = open(str(filename),"w")
        except:
            # Pass errors up to the calling subprogram
            self.__error = "write: error opening file \"%s\"" % str(filename)
            raise
        # Write the header components first
        #
        # Application name and version
        if str(application) != "":
            f.write("#CCP4I VERSION "+str(application))
            if str(version) != "":
                f.write(" "+str(version)+"\n")
        else:
            f.write("#CCP4I VERSION "+str(self.__application)+"\n")
        # Def file identifier
        if str(defname) != "":
            defid = str(defname)
        else:
            # Not supplied; use the base of the filename
            basename = os.path.splitext(os.path.basename(filename))
            if len(basename) > 0:
                defid = basename[0]
        f.write("#CCP4I SCRIPT DEF "+str(defid)+"\n")
        # Date
        f.write("#CCP4I DATE "+GetDefDate()+"\n")
        # Username
        if str(user) != "":
            userid = str(user)
        else:
            userid = str(GetUserId())
        f.write("#CCP4I USER "+str(userid)+"\n")
        # Project name
        if str(project) != "":
            f.write("#CCP4I PROJECT "+str(project)+"\n")

        # Write the actual data
        for prototype in self.__parameters:
            if not self.parameter_isindexed(prototype):
                self.writeparameter(prototype,f,include_types)
            else:
                for parameter in self.listindexed(prototype,
                                                  include_prototype=include_prototypes):
                    self.writeparameter(parameter,f,include_types)
        # Finished
        f.close()

    def writeparameter(self,parameter,f,include_types=False):
        """Write an array parameter to a def file.

        The calling function must supply the complete parameter name (i.e.
        including index, if appropriate) and an initialised file object f.
        If include_types is True then the parameter type is also written."""
        
        f.write(str(parameter)+"\t")
        if include_types:
            f.write(str(self.typeof(parameter))+"\t")
        value = str(self[parameter])
        if len(value) == 0:
            # Empty string
            f.write("\"\"\n")
        elif value.count(" ") > 0:
            # Contains spaces
            f.write("\""+value+"\"\n")
        else:
            # Single nonblank value
            f.write(value+"\n")
        return

    # Methods for extracting data from the header
    def getheaderdata(self,pattern):
        # Given a pattern, find if there is a header
        # line with this data and return the line
        rx = re.compile(pattern)
        for line in self.__header:
            if rx.match(line):
                return line
        return ""

    # Methods for dealing with parameters
    # An array parameter can be single valued or indexed
    # The "prototype" of an indexed parameter is PARAM,0

    def parameters(self):
        # Returns a list of the parameters in the array
        return copy.copy(self.__parameters)

    def has_parameter(self,parameter,*indices):
        # Check for the existence of a parameter
        # Optionally up to two indices can be specified
        if len(indices) > 2:
            raise KeyError, "Too many indices for has_parameter"
        key = str(parameter)
        if len(indices) > 0:
            key = key+","+str(indices[0])
            if len(indices) > 1:
                key = key+"_"+str(indices[1])
        return self.__contents.has_key(key)

    def addparameter(self,parameter,index="",type="",value=""):
        # Add a new parameter to the array object
        # Optionally specify:
        # index, type and (default) value the parameter
        if index == "":
            key = self.__parameter_prototype(parameter)
        else:
            key = self.__parameter_prototype(
                self.__make_parameter_name(parameter,index))
        if not self.isparameter(key):
            self.__parameters.append(key)
            self.__types[key] = type
            self[key] = value
            return True
        else:
            return False

    def delparameter(self,parameter,*indices):
        # Delete a parameter from the array object
        # Optionally up to two indices can be specified
        if len(indices) > 2:
            raise KeyError, "Too many indices for has_parameter"
        key = str(parameter)
        if len(indices) > 0:
            key = key+","+str(indices[0])
            if len(indices) > 1:
                key = key+"_"+str(indices[1])
        del self.__contents[key]

    def isparameter(self,parameter):
        # Check whether the parameter is "prototyped" in
        # the list of parameters
        key = self.__parameter_prototype(parameter)
        if self.__parameters.count(key) > 0:
            return True
        else:
            return False

    def typeof(self,parameter):
        # Return the type for a particular parameter
        key = self.__parameter_prototype(parameter)
        try:
            return self.__types[key]
        except:
            self.__error = "No type for parameter "+str(parameter)
            return ""

    def listindexed(self,parameter,include_prototype=False):
        # Given the name of parameter, return a list of
        # all the parameters which share the same prototype
        # If include_prototype is True then the prototype parameter
        # name is also included in the list
        key = self.__parameter_prototype(parameter)
        parameters = []
        for p in self.__contents.keys():
            if p == self.__parameter_prototype(p) and not include_prototype:
                continue
            if key == self.__parameter_prototype(p):
                parameters.append(p)
        # If no results were returned and the parameter
        # was not indexed then try it again with the zero index
        if len(parameters) == 0 and parameter.find(",") < 0:
            key = parameter+",0"
            parameters = self.listindexed(key,include_prototype)
        return parameters

    def getequivalent(self,parameter1,parameter2):
        # Given an indexed parameter "parameter1", return
        # the name of a parameter with the same index but
        # based on "parameter2"
        index = self.__parameter_index(parameter1)
        if index != "":
            return self.__parameter_root(parameter2)+","+str(index)
        else:
            # No index on the first parameter
            return parameter2

    def getvalue(self,parameter,*indices):
        # Fetch the value of a parameter
        # Up to two indices can be specified
        if len(indices) > 2:
            raise KeyError, "Too many indices for getvalue"
        key = str(parameter)
        if len(indices) > 0:
            key = key+","+str(indices[0])
            if len(indices) > 1:
                key = key+"_"+str(indices[1])
        return self[key]

    def setvalue(self,parameter,*indices):
        # Assign a value to a parameter
        # Up to two indices can be specified
        # The last argument must always be a value to be assigned -
        # if this is None then the default will be used
        if len(indices) > 3:
            raise KeyError, "Too many indices for setvalue"
        if len(indices) == 2:
            key = self.__make_parameter_name(parameter,indices[0])
        elif len(indices) == 3:
            key = self.__make_parameter_name(parameter,indices[0],indices[1])
        else:
            key = str(parameter)
        value = indices[-1]
        if value != None:
            self[key] = value
        else:
            self[key] = self[self.__parameter_prototype(key)]

    def newvalue(self,parameter,*indices):
        # Make a new occurance of an indexed parameter and
        # assign a value to it
        # Only a single index can be specified
        # The last argument must be the value to be assigned -
        # if this is None then the default will be used
        # The new index is returned
        n = len(indices) - 1
        if n < 0 or n > 1:
            raise KeyError, "Wrong number of indices for newvalue"
        key = str(parameter)
        if n == 0:
            # Check that the parameter is not already indexed
            if key.find(",") > 0:
                n = 1
        else:
            key = key+","+str(indices[0])
        print "newvalue: key = "+key
        # Find all the occurances of the indexed parameter
        # and get the current maximum
        max = 0
        for p in self.__contents.keys():
            if str(p).startswith(key):
                #print "Matched "+p+" with "+key
                if n == 1:
                    i = p.rfind("_")
                else:
                    i = p.find(",")
                #print "Trailing index is "+str(p[i+1:])
                if int(p[i+1:]) > max:
                    max = int(p[i+1:])
        #print "newvalue: highest current index = ",str(max)
        # Make the next indexed parameter
        max += 1
        if n == 0:
            key = key+","+str(max)
        else:
            key = key+"_"+str(max)
        #print "New parameter = "+key
        value = indices[-1]
        if value != None:
            self[key] = value
        else:
            self[key] = self[self.__parameter_prototype(key)]
        return max

    def incr(self,parameter,increment=1):
        # Alter the value of a numerical parameter
        # Increment can be positive or negative
        try:
            self[parameter] = int(self[parameter]) + increment
            return int(self[parameter])
        except:
            raise TypeError, "incr operation failed"

    def replacevalue(self,parameter,*indices):
        # Replace the value of one indexed parameter
        # with that of the same parameter at a different
        # index position
        # Also deletes the second indexed parameter
        if len(indices) == 2:
            key1 = self.__make_parameter_name(parameter,indices[0])
            key2 = self.__make_parameter_name(parameter,indices[1])
        elif len(indices) == 4:
            key1 = self.__make_parameter_name(parameter,indices[0],indices[1])
            key2 = self.__make_parameter_name(parameter,indices[2],indices[3])
        else:
            raise KeyError, "Wrong number of indices for replacevalue"
        self[key1] = self[key2]
        del self[key2]

    def parameter_root(self,parameter):
        # Wraps __parameter_root
        return self.__parameter_root(parameter)

    def parameter_index(self,parameter):
        # Wraps __parameter_index()
        return self.__parameter_index(parameter)

    def parameter_isindexed(self,parameter):
        # Wraps __parameter_isindexed()
        return self.__parameter_isindexed(parameter)

    # Internal (private) methods

    def __parameter_isindexed(self,parameter):
        # Private: determine if a def file parameter is an
        # indexed parameter
        comma = str(parameter).find(",")
        if comma > -1:
            return True
        else:
            return False

    def __parameter_root(self,parameter):
        # Private: return the parameter name stripped
        # of its index
        comma = str(parameter).find(",")
        if comma > -1:
            return str(parameter)[:comma]
        else:
            return str(parameter)

    def __parameter_index(self,parameter):
        # Private: return the index part of a parameter name
        comma = str(parameter).find(",")
        if comma > -1:
            return str(parameter)[comma+1:]
        else:
            return ""

    def __parameter_prototype(self,parameter):
        # Private: return the prototype name for the given
        # parameter
        root = self.__parameter_root(parameter)
        if parameter == root:
            return root
        else:
            return root+",0"

    def __make_parameter_name(self,parameter,index1="",index2=""):
        # Private: construct the name for an indexed parameter
        key = parameter
        if index1 != "":
            key = str(key)+","+str(index1)
            if index2 != "":
                key = str(key)+"_"+str(index2)
        return key

#######################################################################
# General supporting functions
#######################################################################

# Return an absolute path
def GetAbsolutePath(filen):
    """Return the absolute path - assume that relative paths are rooted
    at the current working directory."""

    return os.path.abspath(filen)

# Make a directory
def MakeDir(directory):
    """Create a new directory."""
    
    return os.mkdir(directory)

# Check if a directory exists
def DirExists(directory):
    """Check for the existence of a directory.

    This will also return a negative result if the target exists
    but is not a directory."""
    
    return os.path.isdir(directory)

# Check if a file exists
def FileExists(filen):
    """Check for the existence of a file."""
    
    return os.path.isfile(filen)

# Get the rootname of a file i.e. drop the trailing extension
def GetFileRootname(filen):
    """Return the rootname of a filename."""

    components = os.path.splitext(filen)
    return components[0]

# Return the user id
def GetUserId():
    """Return the username of the current user.

    This function returns the value of the environment variable
    USER, and may not work on systems that do not set this variable."""
    return os.environ.get("USER")

# Return the current date and time
def GetDate():
    """Return the current date.

    The date is in the format e.g. 27 Feb 2006  15:05:34."""
    
    return time.strftime("%d %b %Y  %H:%M:%S")

# Return the current data and time (in def file format)
def GetDefDate():
    """Return the current date for a def file header.

    The date is in the format dd/mm/yy HH:MM:SS."""
    
    return time.strftime("%d/%m/%y %H:%M:%S")

# Return the current process id
def GetPid():
    """Return the current process id."""
    
    return os.getpid()

#######################################################################
# Handling the platform and the CCP4 environment
#######################################################################

def GetPlatform():
    """Return the platform information."""

    return sys.platform

def GetOPSYS():
    """Return 'WINDOWS' or 'UNIX', depending on the operating system."""

    Platform = GetPlatform();
    unix = re.compile(r'^linux|irix|osf1|sunos|darwin')
    windows = re.compile(r'^win')
    if unix.match(Platform):
        return "UNIX"
    elif windows.match(Platform):
        return "WINDOWS"
    else:
        return "UNIX"

def GetOpsys():
    """Return 'windows' or 'unix', depending on the operating system."""

    Platform = GetOPSYS()
    return Platform.lower()

def SearchPath(topname,*elements):
    """Return the full path name for a CCP4i code or data file."""

    # This is loosely based on proc SearchPath in system.tcl
    # topname should be either TOP or HELP

    if topname == "TOP":
        # First look in the user's .CCP4 area
        filen = os.path.join(GetDotCCP4(),"CCP4I_TOP")
        for ele in elements:
            filen = os.path.join(filen,str(ele))
        if FileExists(filen):
            return filen

    # Otherwise return the path in the main CCP4i area
    filen = os.path.join(GetEnvVar("CCP4I_"+str(topname)))
    for ele in elements:
        filen = os.path.join(filen,str(ele))
    return filen

def GetDotCCP4():
    """Return the location of the .CCP4 directory."""

    # This encodes the logic found in system.tcl, proc InitialiseDotCCP4
    if GetOPSYS() == "UNIX":
        return os.path.join(GetEnvVar("HOME"),".CCP4")
    elif GetOPSYS() == "WINDOWS":
        if GetEnvVar("USERPROFILE") == "":
            return os.path.join(GetEnvVar("SystemRoot"),"Profiles",LoginName,"CCP4")
        else:
            return os.path.join(GetEnvVar("USERPROFILE"),"CCP4")
    # We shouldn't reach this point so it should probably raise an exception
    return ""

def GetEnvVar(var):
    """Return the value of an environment variable."""

    if os.environ.has_key(var):
        return os.environ[var]
    else:
        return ""

#######################################################################
# Supporting database handling functions
#######################################################################

# Return the CCP4 database directory given the project
# directory
def GetDbDir(projectdir):
    """Return the name of the CCP4i database directory for a project.
    
    Given the project directory path, the name of the CCP4i database
    subdircetory is returned."""
    
    return os.path.join(projectdir,"CCP4_DATABASE")

# Return the CCP4 database.def file name given the
# project directory
def GetDbFile(projectdir):
    """Return the name of the database.def file.
    
    The sole argument is the name of the project directory."""
    
    return os.path.join(GetDbDir(projectdir),"database.def")

# Make a new CCP4i database.def file
# This consists of the number of jobs plus a dummy record zero
def MakeDbFile(dbfile,project):
    """Make a new CCP4i database.def file.

    The arguments are the absolute filename and the name of the
    project. A database file will be generated from the template
    database.def file in the CCP4 distribution."""

    dbarray = array(initfile=SearchPath("TOP","etc","database.def"))
    dbarray.write(dbfile,project=project)

# Return a list of the items in a database record
def ListDbItems():
    """Return a template list of the items in database record."""
    
    return ["TITLE", \
            "TASKNAME", \
            "DATE", \
            "STATUS", \
            "INPUT_FILES", \
            "INPUT_FILES_DIR", \
            "INPUT_FILES_STATUS", \
            "OUTPUT_FILES", \
            "OUTPUT_FILES_DIR", \
            "OUTPUT_FILES_STATUS", \
            "LOGFILE"]

#######################################################################
# CCP4i def file supporting functions
#######################################################################

# Tokenise a line from a def file
def tokenise(line):
    """Tokenise a string and return a list.

    Typically the string will be a line from a CCP4i def file. Tokens
    are delimited by whitespace and by double quotes, with quoted whitespace
    within a token not counting as a delimiter."""
    
    sline = str(line)
    tokens = []
    token = False
    quote = False
    start = 0
    for i in range(len(sline)):
        c = sline[i]
        if token and not quote:
            if c == " " or c == "\t":
                # end of current token
                tokens.append(sline[start:i])
                token = False
                quote = False
        elif token and c == "\"":
                # Detected a quote - flip the quote flag
                if quote:
                    quote = False
                else:
                    quote = True
        elif not token:
            if c != " " and c != "\t":
                # Start of a new token
                token = True
                start = i
                if c == "\"":
                    # Also it's quoted
                    quote = True

    # End of the loop
    if token:
        # End of the last token
        tokens.append(sline[start:len(sline)])
    return tokens

#######################################################################
# Other supporting functions
#######################################################################

# Return the ccp4i module version number
def version():
    """Return the version number of the module extracted from the RCS string."""

    # Note: need to break up the regular expression pattern otherwise
    # CVS will substitute it on commit
    pattern = re.compile(r'^\$'+'Revision: ([0-9\.]+) \$') 
    if pattern.match(__version__):
        return str(pattern.match(__version__).group(1))
    else:
        return "?"

