#############################################################
#
# ccp4icompat_client.tcl
#
# Client application that connects to the handler, opens a
# project and then:
# 1. overrides the lock file
# 2. changes the database.def file modification time
# in order to test compatibility with CCP4i.
#
# CVS_Id $Id: ccp4icompat_client.tcl,v 1.1 2007/02/09 15:51:46 pjx Exp $
#
#############################################################

# Announce yourself
puts "=========================================================="
puts " ccp4icompat_client\n"
puts " Connect to the handler, open a project and then:"
puts " 1. override the lock file"
puts " 2. changes the database.def file modification time"
puts " in order to test compatibility with CCP4i."
puts "==========================================================\n"

# Acquire the handler client API functions
source [file join $env(BIOXHIT_TOP) dbccp4i ClientAPI dbClientAPI.tcl]

# Get command line argument
if { [llength $argv] == 1 } {
    set project [lindex $argv 0]
    puts "Making fake lock on database in project \"$project\"\n"
} else {
    puts "Usage: tclsh ccp4icompat_client.tcl <project_name>"
    exit 0
}

# Set the debugging output level
DbXMLSetDebug 0

# Start the handler
if { ![DbStartHandler] } {
    puts "Test failed: couldn't start the handler"
    exit 1
}

# Connect to the handler
if { ![DbHandlerConnect] } {
    puts "Test failed: couldn't connect to the handler"
    exit 1
}

# List the available projects
set projects [ListProjects]
puts "Projects: $projects"
if { [lsearch $projects $project] < 0 } {
    puts "Test failed: can't find project called \"$project\""
    DbHandlerDisconnect
    exit 1
}

# Test for an existing lock on the project
set projectdir [GetProjectDir $project]
set lockfile [file join $projectdir CCP4_DATABASE database.LOCK]
if { [file exists $lockfile] } {
    puts "Test failed: $project already has a lockfile"
    DbHandlerDisconnect
    exit 1
}

# Attempt to open the database
puts "----------------------------------------------------------"
puts " Opening project database..."
puts "----------------------------------------------------------"
if { [OpenDatabase $project msg] } {
    puts "Test succeeded: open succeeded"
} else {
    puts "Test failed: open failed ($msg)\n"
    if { [file exists $lockfile] } {
	file delete $lockfile
    }
    DbHandlerDisconnect
    exit 1
}

# Try to obtain and verify data
set njobs [GetNJOB $project]
puts "NJOBS: $njobs\n"
if { $njobs < 0 } {
    # NJOBS should be zero or greater in an open project
    puts "Test failed: njobs is less than zero"
} else {
    set joblist [ListJobs $project]
    puts "Joblist: \"$joblist\" (number of jobs = [llength $joblist])\n"
    if { $njobs > 0 && [llength $joblist] == 0 } {
	puts "Test encountered inconsistency in data!"
    }
}

# Simulate CCP4i overriding the lock by writing a new file
puts "----------------------------------------------------------"
puts " Overwriting lockfile in project..."
puts "----------------------------------------------------------"
file delete $lockfile
set fp [open $lockfile "w"]
puts $fp "Lock file for $project"
puts $fp "Created by nobody"
puts $fp "Date 31 Oct 2006  16:02:27"
puts $fp "Owned by pid 0"
close $fp

# Try to obtain and verify data once again
set njobs1 [GetNJOB $project]
puts "NJOBS(1): $njobs1\n"
if { $njobs != $njobs1 } {
    # This should agree with the previous value
    puts "Test failed: njobs should agree with the previous value"
} else {
    set joblist [ListJobs $project]
    puts "Joblist: \"$joblist\" (number of jobs = [llength $joblist])\n"
    if { $njobs1 > 0 && [llength $joblist] == 0 } {
	puts "Test encountered inconsistency in data!"
    }
}

# Simulate CCP4i overwriting database by rewriting the database file
puts "----------------------------------------------------------"
puts " Updating database.def file in project..."
puts "----------------------------------------------------------"
set db_file [file join $projectdir CCP4_DATABASE database.def]
puts "$db_file: mtime = [file mtime $db_file]"
set db_copy "[subst $db_file].copy"
after 1000
set fdf [open $db_file "r"]
set fcp [open $db_copy "w"]
while { [gets $fdf line] >= 0 } {
    puts $fcp $line
}
close $fdf
close $fcp
file rename -force $db_copy $db_file
puts "$db_file: mtime = [file mtime $db_file]"

# Try to obtain and verify data once again
set njobs2 [GetNJOB $project]
puts "NJOBS(2): $njobs2\n"
if { $njobs2 != 0 && $njobs != 0 } {
    # NJOBS should be zero if data is no longer readable
    puts "Test failed: njobs is not zero"
} else {
    set joblist [ListJobs $project]
    puts "Joblist: \"$joblist\" (number of jobs = [llength $joblist])\n"
    if { $njobs2 > 0 && [llength $joblist] == 0 } {
	puts "Test encountered inconsistency in data!"
    }
}

# Finished
CloseDatabase $project
DbHandlerDisconnect
if { [file exists $lockfile] } {
    file delete $lockfile
}
puts "=========================================================="
puts " ccp4icompat_client: finished"
puts "=========================================================="
exit 0